// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Driver for Broadcom MPI3 Storage Controllers
 *
 * Copyright (C) 2017-2022 Broadcom Inc.
 *  (mailto: mpi3mr-linuxdrv.pdl@broadcom.com)
 *
 */

#include "mpi3mr.h"
#include "mpi3mr_app.h"

static struct fasync_struct *mpi3mr_app_async_queue;

/* SysFS task management type definitions*/
enum mpi3mr_sysfs_tm {
	MPI3MR_SYSFS_TM_SOFT_RESET = 1,
	MPI3MR_SYSFS_TM_ABORT_TASK = 3,
	MPI3MR_SYSFS_TM_TARGET_RESET,
	MPI3MR_SYSFS_TM_LUN_RESET,
	MPI3MR_SYSFS_TM_ABORT_TASK_SET,
	MPI3MR_SYSFS_TM_DIAG_FAULT_RESET,
};

#define MPI3MR_SYSFS_TM_TIMEOUT		120 /*in seconds*/

/* Encapsulated NVMe command definitions */
#define	MPI3MR_NVME_PRP_SIZE		8 /* PRP size */
#define	MPI3MR_NVME_CMD_PRP1_OFFSET	24 /* PRP1 offset in NVMe cmd */
#define	MPI3MR_NVME_CMD_PRP2_OFFSET	32 /* PRP2 offset in NVMe cmd */
#define	MPI3MR_NVME_CMD_SGL_OFFSET	24 /* SGL offset in NVMe cmd */
#define MPI3MR_NVME_DATA_FORMAT_PRP	0
#define MPI3MR_NVME_DATA_FORMAT_SGL1	1
#define MPI3MR_NVME_DATA_FORMAT_SGL2	2

/**
 * struct mpi3mr_nvme_pt_sge -  Structure to store SGEs for NVMe
 * Encapsulated commands.
 *
 * @base_addr: Physical address
 * @length: SGE length
 * @rsvd: Reserved
 * @rsvd1: Reserved
 * @sgl_type: sgl type
 */
struct mpi3mr_nvme_pt_sge {
	u64 base_addr;
	u32 length;
	u16 rsvd;
	u8 rsvd1;
	u8 sgl_type;
};

/* Data direction definitions*/
#define DMA_BIDIRECTIONAL	0
#define DMA_TO_DEVICE	1
#define DMA_FROM_DEVICE	2

/**
 * struct mpi3mr_buf_map -  local structure to
 * track kernel and user buffers associated with an IOCTL
 * structure.
 *
 * @user_buf: User buffer virtual address
 * @kern_buf: Kernel buffer virtual address
 * @kern_buf_dma: Kernel buffer DMA address
 * @user_buf_len: User buffer length
 * @kern_buf_len: Kernel buffer length
 * @data_dir: Data direction.
 */
struct mpi3mr_buf_map {
	void __user *user_buf;
	void *kern_buf;
	dma_addr_t kern_buf_dma;
	u32 user_buf_len;
	u32 kern_buf_len;
	u8 data_dir;
};

/**
 * mpi3mr_diag_buffer_for_type - returns buffer desc for type
 * @mrioc: Adapter instance reference
 * @buffer_type: Diagnostic buffer type
 *
 * Identifies matching diag descriptor from mrioc for given diag
 * buffer type.
 *
 * Return: diag buffer descriptor on success, NULL on failures.
 */

static inline struct diag_buffer_desc *
mpi3mr_diag_buffer_for_type(struct mpi3mr_ioc *mrioc, u8 buf_type)
{
	u8 i;

	for (i = 0; i < MPI3MR_MAX_NUM_HDB; i++) {
		if (mrioc->diag_buffers[i].type == buf_type)
			return &mrioc->diag_buffers[i];
	}
	return NULL;
}

/**
 * mpi3mr_hdbstatuschg_evt_th - HDB status change evt tophalf
 * @mrioc: Adapter instance reference
 * @event_reply: event data
 *
 * Modifies the status of the applicable diag buffer descriptors
 *
 * Return: Nothing
 */
void mpi3mr_hdbstatuschg_evt_th(struct mpi3mr_ioc *mrioc,
	struct mpi3_event_notification_reply *event_reply)
{
	struct mpi3_event_data_diag_buffer_status_change *evtdata =
	    (struct mpi3_event_data_diag_buffer_status_change *)event_reply->event_data;
	struct diag_buffer_desc *diag_buffer;

	diag_buffer = mpi3mr_diag_buffer_for_type(mrioc, evtdata->type);
	if (!diag_buffer)
		return;
	if ((diag_buffer->status != MPI3MR_HDB_BUFSTATUS_POSTED_UNPAUSED) &&
	    (diag_buffer->status != MPI3MR_HDB_BUFSTATUS_POSTED_PAUSED))
		return;
	switch (evtdata->reason_code) {
	case MPI3_EVENT_DIAG_BUFFER_STATUS_CHANGE_RC_RELEASED:
	{
		diag_buffer->status = MPI3MR_HDB_BUFSTATUS_RELEASED;
		mpi3mr_app_send_aen(mrioc);
		break;
	}
	case MPI3_EVENT_DIAG_BUFFER_STATUS_CHANGE_RC_RESUMED:
	{
		diag_buffer->status = MPI3MR_HDB_BUFSTATUS_POSTED_UNPAUSED;
		break;
	}
	case MPI3_EVENT_DIAG_BUFFER_STATUS_CHANGE_RC_PAUSED:
	{
		diag_buffer->status = MPI3MR_HDB_BUFSTATUS_POSTED_PAUSED;
		break;
	}
	default:
		dprint_event_th(mrioc, "%s: unknown reason_code(%d)\n",
		    __func__, evtdata->reason_code);
		break;
	}
}

/**
 * mpi3mr_alloc_diag_bufs - Allocate memory for diag buffers
 * @mrioc: Adapter instance reference
 *
 * This functions checks whether the driver defined buffer sizes
 * are greater than IOCFacts provided controller local buffer
 * sizes and if the driver defined sizes are more then the
 * driver allocates the specific buffer by reading driver page1
 *
 * Return: Nothing.
 */
void mpi3mr_alloc_diag_bufs(struct mpi3mr_ioc *mrioc)
{
	struct diag_buffer_desc *diag_buffer;
	struct mpi3_driver_page1 driver_pg1;
	u32 trace_dec_size, trace_min_size, fw_dec_size, fw_min_size,
		trace_size, fw_size;
	u16 pg_sz = sizeof(driver_pg1);
	int retval = 0;

	if (mrioc->diag_buffers[0].addr || mrioc->diag_buffers[1].addr)
		return;

	retval = mpi3mr_cfg_get_driver_pg1(mrioc, &driver_pg1, pg_sz);
	if (retval) {
		ioc_warn(mrioc, "%s: driver page 1 read failed, allocating "
			   "default trace/fw diag buffer sizes\n", __func__);
		trace_size = fw_size = MPI3MR_DEFAULT_HDB_MAX_SZ;
		trace_dec_size = fw_dec_size = MPI3MR_DEFAULT_HDB_DEC_SZ;
		trace_min_size = fw_min_size = MPI3MR_DEFAULT_HDB_MIN_SZ;

	} else {
		trace_size = driver_pg1.host_diag_trace_max_size * 1024;
		trace_dec_size = driver_pg1.host_diag_trace_decrement_size
			 * 1024;
		trace_min_size = driver_pg1.host_diag_trace_min_size * 1024;
		fw_size = driver_pg1.host_diag_fw_max_size * 1024;
		fw_dec_size = driver_pg1.host_diag_fw_decrement_size * 1024;
		fw_min_size = driver_pg1.host_diag_fw_min_size * 1024;
		if ((trace_size == 0) && (fw_size == 0)) {
			dprint_init(mrioc, "%s:Invalid buffer sizes read from "
					   "driver page1 tracesize = %dKB,"
					   "fwsize = %dKB\n",
					__func__, trace_size, fw_size);
			return;
		}
	}

retry_trace:

	diag_buffer = &mrioc->diag_buffers[0];
	diag_buffer->type = MPI3_DIAG_BUFFER_TYPE_TRACE;
	diag_buffer->status = MPI3MR_HDB_BUFSTATUS_NOT_ALLOCATED;
	if ((mrioc->facts.diag_trace_sz < trace_size) && (trace_size >=
		trace_min_size)) {
		diag_buffer->addr = dma_zalloc_coherent(&mrioc->pdev->dev,
		    trace_size, &diag_buffer->dma_addr, GFP_KERNEL);
		if (diag_buffer->addr) {
			dprint_init(mrioc, "%s: host diag trace memory "
			    "allocated = %dKB\n", __func__, trace_size / 1024);
			diag_buffer->size = trace_size;
		} else {
			trace_size -= trace_dec_size;
			goto retry_trace;
		}
	}

retry_fw:

	diag_buffer = &mrioc->diag_buffers[1];

	diag_buffer->type = MPI3_DIAG_BUFFER_TYPE_FW;
	diag_buffer->status = MPI3MR_HDB_BUFSTATUS_NOT_ALLOCATED;
	if ((mrioc->facts.diag_fw_sz < fw_size) && (fw_size >= fw_min_size)) {
		diag_buffer->addr = dma_zalloc_coherent(&mrioc->pdev->dev,
		    fw_size, &diag_buffer->dma_addr, GFP_KERNEL);
		if (diag_buffer->addr) {
			dprint_init(mrioc, "%s: host diag fw memory "
			    "allocated = %dKB\n", __func__, fw_size / 1024);
			diag_buffer->size = fw_size;
		} else {
			fw_size -= fw_dec_size;
			goto retry_fw;
		}
	}
}

/**
 * mpi3mr_issue_diag_buf_post - Send diag buffer post req
 * @mrioc: Adapter instance reference
 * @diag_buffer: Diagnostic buffer descriptor
 *
 * Issue diagnostic buffer post MPI request through admin queue
 * and wait for the completion of it or time out.
 *
 * Return: 0 on success, non-zero on failures.
 */

static int mpi3mr_issue_diag_buf_post(struct mpi3mr_ioc *mrioc,
    struct diag_buffer_desc *diag_buffer)
{
	struct mpi3_diag_buffer_post_request diag_buf_post_req;
	u8 prev_status;
	int retval = 0;

	memset(&diag_buf_post_req, 0, sizeof(diag_buf_post_req));
	mutex_lock(&mrioc->init_cmds.mutex);
	if (mrioc->init_cmds.state & MPI3MR_CMD_PENDING) {
		dprint_ioctl_err(mrioc, "%s: command is in use\n", __func__);
		mutex_unlock(&mrioc->init_cmds.mutex);
		return -1;
	}
	mrioc->init_cmds.state = MPI3MR_CMD_PENDING;
	mrioc->init_cmds.is_waiting = 1;
	mrioc->init_cmds.callback = NULL;
	diag_buf_post_req.host_tag = cpu_to_le16(MPI3MR_HOSTTAG_INITCMDS);
	diag_buf_post_req.function = MPI3_FUNCTION_DIAG_BUFFER_POST;
	diag_buf_post_req.type = diag_buffer->type;
	diag_buf_post_req.address = le64_to_cpu(diag_buffer->dma_addr);
	diag_buf_post_req.length = le32_to_cpu(diag_buffer->size);

	dprint_ioctl_info(mrioc, "%s: posting diag buffer type %d\n", __func__,
	    diag_buffer->type);
	prev_status = diag_buffer->status;
	diag_buffer->status = MPI3MR_HDB_BUFSTATUS_POSTED_UNPAUSED;
	init_completion(&mrioc->init_cmds.done);
	retval = mpi3mr_admin_request_post(mrioc, &diag_buf_post_req,
	    sizeof(diag_buf_post_req), 1);
	if (retval) {
		dprint_ioctl_err(mrioc, "%s: admin request post failed\n",
		    __func__);
		goto out_unlock;
	}
	wait_for_completion_timeout(&mrioc->init_cmds.done,
	    (MPI3MR_INTADMCMD_TIMEOUT * HZ));
	if (!(mrioc->init_cmds.state & MPI3MR_CMD_COMPLETE)) {
		mrioc->init_cmds.is_waiting = 0;
		dprint_ioctl_err(mrioc, "%s: command timedout\n", __func__);
		mpi3mr_check_rh_fault_ioc(mrioc,
		    MPI3MR_RESET_FROM_DIAG_BUFFER_POST_TIMEOUT);
		retval = -1;
		goto out_unlock;
	}
	if ((mrioc->init_cmds.ioc_status & MPI3_IOCSTATUS_STATUS_MASK)
	    != MPI3_IOCSTATUS_SUCCESS) {
		dprint_ioctl_err(mrioc,
		    "%s: command failed, buffer_type (%d) ioc_status(0x%04x) log_info(0x%08x)\n",
		    __func__, diag_buffer->type,
		    (mrioc->init_cmds.ioc_status & MPI3_IOCSTATUS_STATUS_MASK),
		    mrioc->init_cmds.ioc_loginfo);
		retval = -1;
		goto out_unlock;
	}
	dprint_ioctl_info(mrioc, "%s: diag buffer type %d posted successfully\n",
	    __func__, diag_buffer->type);

out_unlock:
	if (retval)
		diag_buffer->status = prev_status;
	mrioc->init_cmds.state = MPI3MR_CMD_NOTUSED;
	mutex_unlock(&mrioc->init_cmds.mutex);
	return retval;
}

/**
 * mpi3mr_post_diag_bufs - Post diag buffers to the controller
 * @mrioc: Adapter instance reference
 *
 * This function calls helper function to post both trace and
 * firmware buffers to the controller.
 *
 * Return: None
 */
int mpi3mr_post_diag_bufs(struct mpi3mr_ioc *mrioc)
{
	u8 i;
	struct diag_buffer_desc *diag_buffer;

	for (i = 0; i < MPI3MR_MAX_NUM_HDB; i++) {
		diag_buffer = &mrioc->diag_buffers[i];
		if (!(diag_buffer->addr))
			continue;
		if (mpi3mr_issue_diag_buf_post(mrioc, diag_buffer))
			return -1;
	}
	return 0;
}

/**
 * mpi3mr_issue_diag_buf_release - Send diag buffer release req
 * @mrioc: Adapter instance reference
 * @diag_buffer: Diagnostic buffer descriptor
 *
 * Issue diagnostic buffer manage MPI request with release
 * action request through admin queue and wait for the
 * completion of it or time out.
 *
 * Return: 0 on success, non-zero on failures.
 */

static int mpi3mr_issue_diag_buf_release(struct mpi3mr_ioc *mrioc,
    struct diag_buffer_desc *diag_buffer)
{
	struct mpi3_diag_buffer_manage_request diag_buf_manage_req;
	int retval = 0;

	if ((diag_buffer->status != MPI3MR_HDB_BUFSTATUS_POSTED_UNPAUSED) &&
	    (diag_buffer->status != MPI3MR_HDB_BUFSTATUS_POSTED_PAUSED))
		return retval;

	memset(&diag_buf_manage_req, 0, sizeof(diag_buf_manage_req));
	mutex_lock(&mrioc->init_cmds.mutex);
	if (mrioc->init_cmds.state & MPI3MR_CMD_PENDING) {
		dprint_reset(mrioc, "%s: command is in use\n", __func__);
		mutex_unlock(&mrioc->init_cmds.mutex);
		return -1;
	}
	mrioc->init_cmds.state = MPI3MR_CMD_PENDING;
	mrioc->init_cmds.is_waiting = 1;
	mrioc->init_cmds.callback = NULL;
	diag_buf_manage_req.host_tag = cpu_to_le16(MPI3MR_HOSTTAG_INITCMDS);
	diag_buf_manage_req.function = MPI3_FUNCTION_DIAG_BUFFER_MANAGE;
	diag_buf_manage_req.type = diag_buffer->type;
	diag_buf_manage_req.action = MPI3_DIAG_BUFFER_ACTION_RELEASE;


	dprint_reset(mrioc, "%s: releasing diag buffer type %d\n", __func__,
	    diag_buffer->type);
	init_completion(&mrioc->init_cmds.done);
	retval = mpi3mr_admin_request_post(mrioc, &diag_buf_manage_req,
	    sizeof(diag_buf_manage_req), 1);
	if (retval) {
		dprint_reset(mrioc, "%s: admin request post failed\n", __func__);
		goto out_unlock;
	}
	wait_for_completion_timeout(&mrioc->init_cmds.done,
	    (MPI3MR_INTADMCMD_TIMEOUT * HZ));
	if (!(mrioc->init_cmds.state & MPI3MR_CMD_COMPLETE)) {
		mrioc->init_cmds.is_waiting = 0;
		dprint_reset(mrioc, "%s: command timedout\n", __func__);
		mpi3mr_check_rh_fault_ioc(mrioc,
		    MPI3MR_RESET_FROM_DIAG_BUFFER_RELEASE_TIMEOUT);
		retval = -1;
		goto out_unlock;
	}
	if ((mrioc->init_cmds.ioc_status & MPI3_IOCSTATUS_STATUS_MASK)
	    != MPI3_IOCSTATUS_SUCCESS) {
		dprint_reset(mrioc,
		    "%s: command failed, buffer_type (%d) ioc_status(0x%04x) log_info(0x%08x)\n",
		    __func__, diag_buffer->type,
		    (mrioc->init_cmds.ioc_status & MPI3_IOCSTATUS_STATUS_MASK),
		    mrioc->init_cmds.ioc_loginfo);
		retval = -1;
		goto out_unlock;
	}
	dprint_reset(mrioc, "%s: diag buffer type %d released successfully\n",
	    __func__, diag_buffer->type);

out_unlock:
	mrioc->init_cmds.state = MPI3MR_CMD_NOTUSED;
	mutex_unlock(&mrioc->init_cmds.mutex);
	return retval;
}

/**
 * mpi3mr_release_diag_bufs - Release diag buffers
 * @mrioc: Adapter instance reference
 * @skip_rel_action: Skip release action and set buffer state
 * This function calls helper function to release both trace and
 * firmware buffers from the controller.
 *
 * Return: None
 */
void mpi3mr_release_diag_bufs(struct mpi3mr_ioc *mrioc, u8 skip_rel_action)
{
	u8 i;
	struct diag_buffer_desc *diag_buffer;

	for (i = 0; i < MPI3MR_MAX_NUM_HDB; i++) {
		diag_buffer = &mrioc->diag_buffers[i];
		if (!(diag_buffer->addr))
			continue;
		if (diag_buffer->status == MPI3MR_HDB_BUFSTATUS_RELEASED)
			continue;
		if (!skip_rel_action)
			mpi3mr_issue_diag_buf_release(mrioc, diag_buffer);
		diag_buffer->status = MPI3MR_HDB_BUFSTATUS_RELEASED;
		mpi3mr_app_send_aen(mrioc);
	}
}

/**
 * mpi3mr_app_pel_abort - sends PEL abort request
 * @mrioc: Adapter instance reference
 *
 * This function sends PEL abort request to the firmware through
 * admin request queue.
 *
 * Return: 0 on success, -1 on failure
 */
static int mpi3mr_app_pel_abort(struct mpi3mr_ioc *mrioc)
{
	struct mpi3_pel_req_action_abort pel_abort_req;
	struct mpi3_pel_reply *pel_reply;
	int retval = 0;
	u16 pe_log_status;

	if (mrioc->reset_in_progress) {
		dprint_ioctl_err(mrioc, "%s: reset in progress\n", __func__);
		return -1;
	}
	if (mrioc->block_ioctls) {
		dprint_ioctl_err(mrioc, "%s: ioctls are blocked\n", __func__);
		return -1;
	}

	memset(&pel_abort_req, 0, sizeof(pel_abort_req));
	mutex_lock(&mrioc->pel_abort_cmd.mutex);
	if (mrioc->pel_abort_cmd.state & MPI3MR_CMD_PENDING) {
		dprint_ioctl_err(mrioc, "%s: command is in use\n", __func__);
		mutex_unlock(&mrioc->pel_abort_cmd.mutex);
		return -1;
	}
	mrioc->pel_abort_cmd.state = MPI3MR_CMD_PENDING;
	mrioc->pel_abort_cmd.is_waiting = 1;
	mrioc->pel_abort_cmd.callback = NULL;
	pel_abort_req.host_tag = cpu_to_le16(MPI3MR_HOSTTAG_PEL_ABORT);
	pel_abort_req.function = MPI3_FUNCTION_PERSISTENT_EVENT_LOG;
	pel_abort_req.action = MPI3_PEL_ACTION_ABORT;
	pel_abort_req.abort_host_tag = cpu_to_le16(MPI3MR_HOSTTAG_PEL_WAIT);

	mrioc->pel_abort_requested = 1;
	init_completion(&mrioc->pel_abort_cmd.done);
	retval = mpi3mr_admin_request_post(mrioc, &pel_abort_req,
	    sizeof(pel_abort_req), 0);
	if (retval) {
		retval = -1;
		dprint_ioctl_err(mrioc, "%s: admin request post failed\n",
		    __func__);
		mrioc->pel_abort_requested = 0;
		goto out_unlock;
	}

	wait_for_completion_timeout(&mrioc->pel_abort_cmd.done,
	    (MPI3MR_INTADMCMD_TIMEOUT * HZ));
	if (!(mrioc->pel_abort_cmd.state & MPI3MR_CMD_COMPLETE)) {
		mrioc->pel_abort_cmd.is_waiting = 0;
		dprint_ioctl_err(mrioc, "%s: command timedout\n", __func__);
		if (!(mrioc->pel_abort_cmd.state & MPI3MR_CMD_RESET))
			mpi3mr_soft_reset_handler(mrioc,
			    MPI3MR_RESET_FROM_PELABORT_TIMEOUT, 1);
		retval = -1;
		goto out_unlock;
	}
	if ((mrioc->pel_abort_cmd.ioc_status & MPI3_IOCSTATUS_STATUS_MASK)
	     != MPI3_IOCSTATUS_SUCCESS) {
		dprint_ioctl_err(mrioc,
		    "%s: command failed, ioc_status(0x%04x) log_info(0x%08x)\n",
		    __func__, (mrioc->pel_abort_cmd.ioc_status &
		    MPI3_IOCSTATUS_STATUS_MASK),
		    mrioc->pel_abort_cmd.ioc_loginfo);
		retval = -1;
		goto out_unlock;
	}
	if (mrioc->pel_abort_cmd.state & MPI3MR_CMD_REPLY_VALID) {
		pel_reply = (struct mpi3_pel_reply *)mrioc->pel_abort_cmd.reply;
		pe_log_status = le16_to_cpu(pel_reply->pe_log_status);
		if (pe_log_status != MPI3_PEL_STATUS_SUCCESS) {
			dprint_ioctl_err(mrioc,
			    "%s: command failed, pel_status(0x%04x)\n",
			    __func__, pe_log_status);
			retval = -1;
		}
	}

out_unlock:
	mrioc->pel_abort_cmd.state = MPI3MR_CMD_NOTUSED;
	mutex_unlock(&mrioc->pel_abort_cmd.mutex);
	return retval;
}

/**
 * mpi3mr_app_verify_adapter - verify adapter number is valid
 * @ioc_number: Adapter number
 * @mriocpp: Pointer to hold per adapter instance
 *
 * This function checks whether given adapter number matches
 * with an adapter id in the driver's list and if so fills
 * pointer to the per adapter instance in mriocpp else set that
 * to NULL.
 *
 * Return: Nothing.
 */
static void mpi3mr_app_verify_adapter(int ioc_number,
    struct mpi3mr_ioc **mriocpp)
{
	struct mpi3mr_ioc *mrioc;

	spin_lock(&mrioc_list_lock);
	list_for_each_entry(mrioc, &mrioc_list, list) {
		if (mrioc->id != ioc_number)
			continue;
		spin_unlock(&mrioc_list_lock);
		*mriocpp = mrioc;
		return;
	}
	spin_unlock(&mrioc_list_lock);
	*mriocpp = NULL;
}


/**
 * mpi3mr_app_refresh_hdb_triggers - Refresh HDB trigger data
 * @mrioc: Adapter instance reference
 * @data_out_buf: User buffer with page type
 * @data_out_sz: length of the user buffer.
 *
 * This function reads the controller trigger config page as
 * defined by the input page type and refreshes the driver's
 * local trigger information structures with the controller's
 * config page data.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_app_refresh_hdb_triggers(struct mpi3mr_ioc *mrioc,
    void __user *data_out_buf, uint32_t data_out_sz)
{
	long rval = -EINVAL;
	struct mpi3mr_ioctl_out_refresh_hdb_triggers refresh_triggers;

	if (data_out_sz != sizeof(refresh_triggers)) {
		dprint_ioctl_err(mrioc, "%s: invalid size argument\n",
		    __func__);
		return rval;
	}

	if (copy_from_user(&refresh_triggers, data_out_buf,
	    sizeof(refresh_triggers))) {
		dprint_ioctl_err(mrioc, "%s: copy from user failed\n",
		    __func__);
		return -EFAULT;
	}
	dprint_ioctl_err(mrioc,
	    "%s: ioctl is not supported yet, page_type %d\n",
	    __func__, refresh_triggers.page_type);
	return rval;
}

/**
 * mpi3mr_app_upload_hdb - Upload a specific HDB to user space
 * @mrioc: Adapter instance reference
 * @karg: IOCTL data payload
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_app_upload_hdb(struct mpi3mr_ioc *mrioc,
    struct mpi3mr_ioctl_drv_cmd *karg)
{
	long rval = -EINVAL;
	struct mpi3mr_ioctl_out_upload_hdb upload_hdb;
	struct diag_buffer_desc *diag_buffer;

	if ((karg->data_out_size != sizeof(upload_hdb))) {
		dprint_ioctl_err(mrioc, "%s: invalid size argument\n",
		    __func__);
		goto out;
	}

	if (copy_from_user(&upload_hdb, karg->data_out_buf,
	    sizeof(upload_hdb))) {
		dprint_ioctl_err(mrioc, "%s: copy from user failed\n",
		    __func__);
		rval = -EFAULT;
		goto out;
	}
	if ((!upload_hdb.length) || (karg->data_in_size != upload_hdb.length)) {
		dprint_ioctl_err(mrioc, "%s: invalid length argument\n",
		    __func__);
		goto out;
	}
	diag_buffer = mpi3mr_diag_buffer_for_type(mrioc, upload_hdb.buf_type);
	if ((!diag_buffer) || (!diag_buffer->addr)) {
		dprint_ioctl_err(mrioc, "%s: invalid buffer type %d\n",
		    __func__, upload_hdb.buf_type);
		goto out;
	}
	if ((diag_buffer->status != MPI3MR_HDB_BUFSTATUS_RELEASED) &&
	    (diag_buffer->status != MPI3MR_HDB_BUFSTATUS_POSTED_PAUSED)) {
		dprint_ioctl_err(mrioc,
		    "%s: invalid buffer status %d for type %d\n",
		    __func__, diag_buffer->status, upload_hdb.buf_type);
		goto out;
	}

	if ((upload_hdb.start_offset + upload_hdb.length) > diag_buffer->size) {
		dprint_ioctl_err(mrioc,
		    "%s: invalid start offset %d, length %d for type %d\n",
		    __func__, upload_hdb.start_offset, upload_hdb.length,
		    upload_hdb.buf_type);
		goto out;
	}

	if (copy_to_user(karg->data_in_buf, (diag_buffer->addr +
	     upload_hdb.start_offset), upload_hdb.length)) {
		dprint_ioctl_err(mrioc, "%s: copy to user failed\n", __func__);
		rval = -EFAULT;
		goto out;
	}
	rval = 0;
out:
	return rval;
}


/**
 * mpi3mr_app_repost_hdb - Re-post HDB
 * @mrioc: Adapter instance reference
 * @data_out_buf: User buffer with buffer type
 * @data_out_sz: length of the user buffer.
 *
 * This function retrieves the HDB descriptor corresponding to a
 * given buffer type and if the HDB is in released status then
 * posts the HDB with the firmware.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_app_repost_hdb(struct mpi3mr_ioc *mrioc,
    void __user *data_out_buf, uint32_t data_out_sz)
{
	long rval = 0;
	struct mpi3mr_ioctl_out_repost_hdb repost_hdb;
	struct diag_buffer_desc *diag_buffer;

	if (data_out_sz != sizeof(repost_hdb)) {
		dprint_ioctl_err(mrioc, "%s: invalid size argument\n",
		    __func__);
		rval = -EINVAL;
		goto out;
	}

	if (copy_from_user(&repost_hdb, data_out_buf,
	    sizeof(repost_hdb))) {
		dprint_ioctl_err(mrioc, "%s: copy from user failed\n",
		    __func__);
		rval = -EFAULT;
		goto out;
	}

	diag_buffer = mpi3mr_diag_buffer_for_type(mrioc, repost_hdb.buf_type);
	if ((!diag_buffer) || (!diag_buffer->addr)) {
		dprint_ioctl_err(mrioc, "%s: invalid buffer type %d\n",
		    __func__, repost_hdb.buf_type);
		rval = -EINVAL;
		goto out;
	}
	if (diag_buffer->status != MPI3MR_HDB_BUFSTATUS_RELEASED) {
		dprint_ioctl_err(mrioc,
		    "%s: invalid buffer status %d for type %d\n",
		    __func__, diag_buffer->status, repost_hdb.buf_type);
		rval = -EINVAL;
		goto out;
	}

	if (mpi3mr_issue_diag_buf_post(mrioc, diag_buffer)) {
		dprint_ioctl_err(mrioc, "%s: post failed for type %d\n",
		    __func__, repost_hdb.buf_type);
		rval = -EFAULT;
		goto out;
	}
out:
	return rval;
}

/**
 * mpi3mr_app_query_hdb - Handler for query HDB IOCTL
 * @mrioc: Adapter instance reference
 * @data_in_buf: User buffer to copy the HDB information
 * @data_in_sz: length of the user buffer.
 *
 * This function prepares and copies the host diagnostic buffer
 * entries to the user buffer.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_app_query_hdb(struct mpi3mr_ioc *mrioc,
    void __user *data_in_buf, uint32_t data_in_sz)
{
	long rval = 0;
	struct mpi3mr_ioctl_in_hdb_status *hbd_status;
	struct mpi3mr_hdb_entry *hbd_status_entry;
	u32 length, min_length;
	u8 i;
	struct diag_buffer_desc *diag_buffer;

	length = (sizeof(*hbd_status) + ((MPI3MR_MAX_NUM_HDB - 1) *
		    sizeof(*hbd_status_entry)));
	hbd_status = kmalloc(length, GFP_KERNEL);
	if (!hbd_status) {
		rval = -ENOMEM;
		goto out;
	}
	hbd_status_entry = &hbd_status->entry[0];

	hbd_status->num_hdb_types = MPI3MR_MAX_NUM_HDB;
	for (i = 0; i < MPI3MR_MAX_NUM_HDB; i++) {
		diag_buffer = &mrioc->diag_buffers[i];
		hbd_status_entry->buf_type = diag_buffer->type;
		hbd_status_entry->status = diag_buffer->status;
		hbd_status_entry->trigger_type =
			MPI3MR_HDB_TRIGGER_TYPE_UNKNOWN;
		hbd_status_entry->trigger_data = 0;
		hbd_status_entry->size = (diag_buffer->size / 1024);
		hbd_status_entry++;
	}

	if (data_in_sz < 4)
	{
		dprint_ioctl_err(mrioc, "%s: invalid size passed\n", __func__);
		rval = -EINVAL;
		goto out;
	}
	min_length = min(data_in_sz, length);
	if (copy_to_user(data_in_buf, hbd_status, min_length)) {
		dprint_ioctl_err(mrioc, "%s: copy to user failed\n", __func__);
		rval = -EFAULT;
	}
out:
	kfree(hbd_status);
	return rval;
}

/**
 * mpi3mr_enable_logdata - Handler for log data enable IOCTL
 * @mrioc: Adapter instance reference
 * @data_in_buf: User buffer to copy the max logdata entry count
 * @data_in_sz: length of the user buffer.
 *
 * This function enables log data caching in the driver if not
 * already enabled and return the maximum number of log data
 * entries that can be cached in the driver.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_enable_logdata(struct mpi3mr_ioc *mrioc,
    void __user *data_in_buf, uint32_t data_in_sz)
{
	long rval = -EINVAL;
	struct mpi3mr_logdata_enable logdata_enable;

	if (mrioc->logdata_buf)
		goto copy_user_data;

	mrioc->logdata_entry_sz =
	    (mrioc->reply_sz - (sizeof(struct mpi3_event_notification_reply) - 4))
	    + MPI3MR_IOCTL_LOGDATA_ENTRY_HEADER_SZ;
	mrioc->logdata_buf_idx = 0;

	mrioc->logdata_buf = kcalloc(MPI3MR_IOCTL_LOGDATA_MAX_ENTRIES,
	    mrioc->logdata_entry_sz, GFP_KERNEL);
	if (!mrioc->logdata_buf)
		return -ENOMEM;

copy_user_data:
	memset(&logdata_enable, 0, sizeof(logdata_enable));
	logdata_enable.max_entries =
	    MPI3MR_IOCTL_LOGDATA_MAX_ENTRIES;
	if (data_in_sz >= sizeof(logdata_enable)) {
		if (copy_to_user(data_in_buf, &logdata_enable,
		    sizeof(logdata_enable))) {
			dprint_ioctl_err(mrioc, "%s: copy to user failed\n",
			    __func__);
			rval = -EFAULT;
		} else
			rval = 0;
	}
	return rval;
}

/**
 * mpi3mr_get_logdata - Handler for get log data  IOCTL
 * @mrioc: Adapter instance reference
 * @data_in_buf: User buffer to copy the logdata entries
 * @data_in_sz: length of the user buffer.
 *
 * This function copies the log data entries to the user buffer
 * when log caching is enabled in the driver.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_get_logdata(struct mpi3mr_ioc *mrioc,
    void __user *data_in_buf, uint32_t data_in_sz)
{
	long rval = -EINVAL;
	u16 num_entries, sz, entry_sz = mrioc->logdata_entry_sz;

	if ((!mrioc->logdata_buf) || (data_in_sz < entry_sz))
		return rval;

	num_entries = data_in_sz / entry_sz;
	if (num_entries > MPI3MR_IOCTL_LOGDATA_MAX_ENTRIES)
		num_entries = MPI3MR_IOCTL_LOGDATA_MAX_ENTRIES;
	sz = num_entries * entry_sz;

	if (copy_to_user(data_in_buf, mrioc->logdata_buf, sz)) {
		dprint_ioctl_err(mrioc, "%s: copy to user failed\n", __func__);
		rval = -EFAULT;
	} else
		rval = 0;

	return rval;
}


/**
 * mpi3mr_app_pel_enable - Handler for PEL enable driver IOCTL
 * @mrioc: Adapter instance reference
 * @data_out_buf: User buffer containing PEL enable data
 * @data_out_sz: length of the user buffer.
 *
 * This function is the handler for PEL enable driver IOCTL.
 * Validates the application given class and locale and if
 * requires aborts the existing PEL wait request and/or issues
 * new PEL wait request to the firmware and returns.
 *
 * Return: 0 on success and proper error codes on failure.
 */
static long mpi3mr_app_pel_enable(struct mpi3mr_ioc *mrioc,
    void __user *data_out_buf, uint32_t data_out_sz)
{
	long rval = -EINVAL;
	struct mpi3mr_ioctl_out_pel_enable pel_enable;
	u8 issue_pel_wait;
	u8 tmp_class;
	u16 tmp_locale;

	if (data_out_sz != sizeof(pel_enable)) {
		dprint_ioctl_err(mrioc, "%s: invalid size argument\n",
		    __func__);
		goto out;
	}

	if (copy_from_user(&pel_enable, data_out_buf, sizeof(pel_enable))) {
		dprint_ioctl_err(mrioc, "%s: copy from user failed\n",
		    __func__);
		rval = -EFAULT;
		goto out;
	}
	if (pel_enable.pel_class > MPI3_PEL_CLASS_FAULT) {
		dprint_ioctl_err(mrioc, "%s: out of range class %d sent\n",
			__func__, pel_enable.pel_class);
		rval = 0;
		goto out;
	}
	if (!mrioc->pel_enabled)
		issue_pel_wait = 1;
	else {
		if ((mrioc->pel_class <= pel_enable.pel_class) &&
		    !((mrioc->pel_locale & pel_enable.pel_locale) ^
		      pel_enable.pel_locale)) {
			issue_pel_wait = 0;
			rval = 0;
		} else {
			pel_enable.pel_locale |= mrioc->pel_locale;

			if (mrioc->pel_class < pel_enable.pel_class)
				pel_enable.pel_class = mrioc->pel_class;

			rval = mpi3mr_app_pel_abort(mrioc);
			if (rval) {
				dprint_ioctl_err(mrioc,
				    "%s: pel_abort failed, status(%ld)\n",
				    __func__, rval);
				goto out;
			}
			issue_pel_wait = 1;
		}
	}
	if (issue_pel_wait) {
		tmp_class = mrioc->pel_class;
		tmp_locale = mrioc->pel_locale;
		mrioc->pel_class = pel_enable.pel_class;
		mrioc->pel_locale = pel_enable.pel_locale;
		mrioc->pel_enabled = 1;
		rval = mpi3mr_pel_get_seqnum_post(mrioc, NULL);
		if (rval) {
			mrioc->pel_class = tmp_class;
			mrioc->pel_locale = tmp_locale;
			mrioc->pel_enabled = 0;
			dprint_ioctl_err(mrioc,
			    "%s: pel get sequence number failed, status(%ld)\n",
			    __func__, rval);
		}
	}

out:
	return rval;
}

/**
 * mpi3mr_get_all_tgt_info - Get all target information
 * @mrioc: Adapter instance reference
 * @data_in_buf: User buffer to copy the target information
 * @data_in_sz: length of the user buffer.
 *
 * This function copies the driver managed target devices device
 * handle, persistent ID, bus ID and taret ID to the user
 * provided buffer for the specific controller. This function
 * also provides the number of devices managed by the driver for
 * the specific controller.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_get_all_tgt_info(struct mpi3mr_ioc *mrioc,
    void __user *data_in_buf, uint32_t data_in_sz)
{
	long rval = -EINVAL;
	u16 num_devices = 0, i = 0;
	u8 get_count = 0;
	unsigned long flags;
	struct mpi3mr_tgt_dev *tgtdev;
	struct mpi3mr_device_map_info *devmap_info = NULL;
	struct mpi3mr_all_tgt_info __user *alltgt_info =
		(struct mpi3mr_all_tgt_info *) data_in_buf;
	uint32_t min_entrylen, kern_entrylen = 0, usr_entrylen;

	if (data_in_sz < sizeof(u32)) {
		dprint_ioctl_err(mrioc, "%s: invalid size argument\n",
		    __func__);
		goto out;
	}

	if (data_in_sz == sizeof(u32))
		get_count = 1;

	if (list_empty(&mrioc->tgtdev_list)) {
		get_count = 1;
		goto copy_usrbuf;
	}

	spin_lock_irqsave(&mrioc->tgtdev_lock, flags);
	list_for_each_entry(tgtdev, &mrioc->tgtdev_list, list)
		num_devices++;
	spin_unlock_irqrestore(&mrioc->tgtdev_lock, flags);

	if (get_count)
		goto copy_usrbuf;
	kern_entrylen = num_devices * sizeof(*devmap_info);
	devmap_info = kmalloc(kern_entrylen, GFP_KERNEL);
	if (!devmap_info) {
		rval = -ENOMEM;
		goto out;
	}
	memset((u8 *)devmap_info, 0xFF, kern_entrylen);
	spin_lock_irqsave(&mrioc->tgtdev_lock, flags);
	list_for_each_entry(tgtdev, &mrioc->tgtdev_list, list) {
		if (i < num_devices) {
			devmap_info[i].handle = tgtdev->dev_handle;
			devmap_info[i].perst_id = tgtdev->perst_id;
			if (tgtdev->host_exposed && tgtdev->starget) {
				devmap_info[i].target_id = tgtdev->starget->id;
				devmap_info[i].bus_id =
				    tgtdev->starget->channel;
			}
			i++;
		}
	}
	num_devices = i;
	spin_unlock_irqrestore(&mrioc->tgtdev_lock, flags);
copy_usrbuf:
	if (copy_to_user(&alltgt_info->num_devices, &num_devices,
	    sizeof(num_devices))) {
		dprint_ioctl_err(mrioc, "%s:%d: copy to user failed\n",
		    __func__, __LINE__);
		rval = -EFAULT;
		goto out;
	}
	usr_entrylen = (data_in_sz - sizeof(u32)) / sizeof(*devmap_info);
	usr_entrylen *= sizeof(*devmap_info);
	min_entrylen = min(usr_entrylen, kern_entrylen);
	if (min_entrylen && (copy_to_user(&alltgt_info->dmi, devmap_info,
	    min_entrylen))) {
		dprint_ioctl_err(mrioc, "%s:%d: copy to user failed\n",
		    __func__, __LINE__);
		rval = -EFAULT;
		goto out;
	}
	rval = 0;
out:
	kfree(devmap_info);
	return rval;
}

/**
 * mpi3mr_get_change_count - Get topology change count
 * @mrioc: Adapter instance reference
 * @data_in_buf: User buffer to copy the change count
 * @data_in_sz: length of the user buffer.
 *
 * This function copies the toplogy change count provided by the
 * driver in events and cached in the driver to the user
 * provided buffer for the specific controller.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_get_change_count(struct mpi3mr_ioc *mrioc,
    void __user *data_in_buf, uint32_t data_in_sz)
{
	long rval = -EINVAL;
	struct mpi3mr_change_count chgcnt;

	memset(&chgcnt, 0, sizeof(chgcnt));
	chgcnt.change_count = mrioc->change_count;
	if (data_in_sz >= sizeof(chgcnt)) {
		if (copy_to_user(data_in_buf, &chgcnt, sizeof(chgcnt))) {
			dprint_ioctl_err(mrioc, "%s: copy to user failed\n",
			    __func__);
			rval = -EFAULT;
		} else
			rval = 0;
	}
	return rval;
}

/**
 * mpi3mr_ioctl_adp_reset - Issue controller reset
 * @mrioc: Adapter instance reference
 * @data_out_buf: User buffer with reset type
 * @data_out_sz: length of the user buffer.
 *
 * This function identifies the user provided reset type and
 * issues approporiate reset to the controller and.wait for that
 * to complete and reinitialize the controller and then returns
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_ioctl_adp_reset(struct mpi3mr_ioc *mrioc,
    void __user *data_out_buf, uint32_t data_out_sz)
{
	long rval = -EINVAL;
	struct mpi3mr_ioctl_adp_reset adpreset;
	u8 save_snapdump;

	if (data_out_sz != sizeof(adpreset)) {
		dprint_ioctl_err(mrioc, "%s: invalid size argument\n",
		    __func__);
		goto out;
	}

	if (copy_from_user(&adpreset, data_out_buf, sizeof(adpreset))) {
		dprint_ioctl_err(mrioc, "%s: copy from user failed\n",
		    __func__);
		rval = -EFAULT;
		goto out;
	}


	switch (adpreset.reset_type) {
	case MPI3MR_IOCTL_ADPRESET_SOFT:
		save_snapdump = 0;
		break;
	case MPI3MR_IOCTL_ADPRESET_DIAG_FAULT:
		save_snapdump = 1;
		break;
	default:
		dprint_ioctl_err(mrioc, "%s: unknown reset_type(%d)\n",
		    __func__, adpreset.reset_type);
		goto out;
	}

	rval = mpi3mr_soft_reset_handler(mrioc, MPI3MR_RESET_FROM_IOCTL,
	    save_snapdump);

	if (rval)
		dprint_ioctl_err(mrioc,
		    "%s: reset handler returned error(%ld) for reset type %d\n",
		    __func__, rval, adpreset.reset_type);

out:
	return rval;
}

/**
 * mpi3mr_app_populate_adpinfo - Get adapter info IOCTL handler
 * @mrioc: Adapter instance reference
 * @data_in_buf: User buffer to hold adapter information
 * @data_in_sz: length of the user buffer.
 *
 * This function provides adapter information for the given
 * controller
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_app_populate_adpinfo(struct mpi3mr_ioc *mrioc,
    void __user *data_in_buf, uint32_t data_in_sz)
{
	long rval = -EINVAL;
	struct mpi3mr_ioctl_in_adpinfo adpinfo;
	enum mpi3mr_iocstate ioc_state;

	memset(&adpinfo, 0, sizeof(adpinfo));
	adpinfo.adp_type = MPI3MR_IOCTL_ADPTYPE_AVGFAMILY;
	adpinfo.pci_dev_id = mrioc->pdev->device;
	adpinfo.pci_dev_hw_rev = mrioc->pdev->revision;
	adpinfo.pci_subsys_dev_id = mrioc->pdev->subsystem_device;
	adpinfo.pci_subsys_ven_id = mrioc->pdev->subsystem_vendor;
	adpinfo.pci_bus = mrioc->pdev->bus->number;
	adpinfo.pci_dev = PCI_SLOT(mrioc->pdev->devfn);
	adpinfo.pci_func = PCI_FUNC(mrioc->pdev->devfn);
	adpinfo.pci_seg_id = pci_domain_nr(mrioc->pdev->bus);
	adpinfo.ioctl_ver = MPI3MR_IOCTL_VERSION;

	ioc_state = mpi3mr_get_iocstate(mrioc);
	if (ioc_state == MRIOC_STATE_UNRECOVERABLE)
		adpinfo.adp_state = MPI3MR_IOCTL_ADPSTATE_UNRECOVERABLE;
	else if ((mrioc->reset_in_progress) || (mrioc->block_ioctls))
		adpinfo.adp_state = MPI3MR_IOCTL_ADPSTATE_IN_RESET;
	else if (ioc_state == MRIOC_STATE_FAULT)
		adpinfo.adp_state = MPI3MR_IOCTL_ADPSTATE_FAULT;
	else
		adpinfo.adp_state = MPI3MR_IOCTL_ADPSTATE_OPERATIONAL;

	memcpy((u8 *)&adpinfo.driver_info, (u8 *)&mrioc->driver_info,
	    sizeof(adpinfo.driver_info));

	if (data_in_sz >= sizeof(adpinfo)) {
		if (copy_to_user(data_in_buf, &adpinfo, sizeof(adpinfo))) {
			dprint_ioctl_err(mrioc, "%s: copy to user failed\n",
			    __func__);
			rval = -EFAULT;
		} else {
			rval = 0;
		}
	}
	return rval;
}

/**
 * mpi3mr_app_process_drv_cmds - Driver IOCTL handler
 * @file: File pointer
 * @arg: User data payload buffer for the IOCTL
 *
 * This function is the top level handler for driver commands,
 * this does basic validation of the buffer and identifies the
 * opcode and switches to correct sub handler.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long
mpi3mr_app_process_drv_cmds(struct file *file, void __user *arg)
{
	long rval = -EINVAL;
	struct mpi3mr_ioc *mrioc = NULL;
	struct mpi3mr_ioctl_drv_cmd karg;

	if (copy_from_user(&karg, arg, sizeof(karg)))
		return -EFAULT;
	mpi3mr_app_verify_adapter(karg.mrioc_id, &mrioc);
	if (!mrioc)
		return -ENODEV;

	if (karg.opcode == MPI3MR_DRVIOCTL_OPCODE_ADPINFO) {
		rval = mpi3mr_app_populate_adpinfo(mrioc, karg.data_in_buf,
		    karg.data_in_size);
		return rval;
	}

	if (file->f_flags & O_NONBLOCK) {
		if (!mutex_trylock(&mrioc->ioctl_cmds.mutex))
			return -EAGAIN;
	} else if (mutex_lock_interruptible(&mrioc->ioctl_cmds.mutex))
		return -ERESTARTSYS;

	switch (karg.opcode) {
	case MPI3MR_DRVIOCTL_OPCODE_ADPRESET:
		rval = mpi3mr_ioctl_adp_reset(mrioc, karg.data_out_buf,
		    karg.data_out_size);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_ALLTGTDEVINFO:
		rval = mpi3mr_get_all_tgt_info(mrioc, karg.data_in_buf,
		    karg.data_in_size);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_GETCHGCNT:
		rval = mpi3mr_get_change_count(mrioc, karg.data_in_buf,
		    karg.data_in_size);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_LOGDATAENABLE:
		rval = mpi3mr_enable_logdata(mrioc, karg.data_in_buf,
		    karg.data_in_size);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_GETLOGDATA:
		rval = mpi3mr_get_logdata(mrioc, karg.data_in_buf,
		    karg.data_in_size);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_PELENABLE:
		rval = mpi3mr_app_pel_enable(mrioc, karg.data_out_buf,
		    karg.data_out_size);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_QUERY_HDB:
		rval = mpi3mr_app_query_hdb(mrioc, karg.data_in_buf,
		    karg.data_in_size);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_REPOST_HDB:
		rval = mpi3mr_app_repost_hdb(mrioc, karg.data_out_buf,
		    karg.data_out_size);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_UPLOAD_HDB:
		rval = mpi3mr_app_upload_hdb(mrioc, &karg);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_REFRESH_HDB_TRIGGERS:
		rval = mpi3mr_app_refresh_hdb_triggers(mrioc, karg.data_out_buf,
		    karg.data_out_size);
		break;
	case MPI3MR_DRVIOCTL_OPCODE_UNKNOWN:
	default:
		dprint_ioctl_err(mrioc,
		    "%s: unsupported driver ioctl opcode %d\n", __func__,
		    karg.opcode);
		break;
	}
	mutex_unlock(&mrioc->ioctl_cmds.mutex);
	return rval;
}

/**
 * mpi3mr_ioctl_build_sgl - SGL construction for MPI IOCTLs
 * @mpi_req: MPI request
 * @sgl_offset: offset to start sgl in the MPI request
 * @dma_buffers: DMA address of the buffers to be placed in sgl
 * @bufcnt: Number of DMA buffers
 * @is_rmc: Does the buffer list has management command buffer
 * @is_rmr: Does the buffer list has management response buffer
 * @num_datasges: Number of data buffers in the list
 *
 * This function places the DMA address of the given buffers in
 * proper format as SGEs in the given MPI request.
 *
 * Return: Nothing
 */
static void mpi3mr_ioctl_build_sgl(u8 *mpi_req, uint32_t sgl_offset,
    struct mpi3mr_buf_map *dma_buffers, u8 bufcnt, u8 is_rmc,
    u8 is_rmr, u8 num_datasges)
{
	u8 *sgl = (mpi_req + sgl_offset), count = 0;
	struct mpi3_mgmt_passthrough_request *rmgmt_req =
	    (struct mpi3_mgmt_passthrough_request *)mpi_req;
	struct mpi3mr_buf_map *dma_buff = dma_buffers;
	u8 sgl_flags, sgl_flags_last;

	sgl_flags = MPI3_SGE_FLAGS_ELEMENT_TYPE_SIMPLE |
		MPI3_SGE_FLAGS_DLAS_SYSTEM | MPI3_SGE_FLAGS_END_OF_BUFFER;
	sgl_flags_last = sgl_flags | MPI3_SGE_FLAGS_END_OF_LIST;

	if (is_rmc) {
		mpi3mr_add_sg_single(&rmgmt_req->command_sgl,
		    sgl_flags_last, dma_buff->kern_buf_len,
		    dma_buff->kern_buf_dma);
		sgl = (u8 *)dma_buff->kern_buf + dma_buff->user_buf_len;
		dma_buff++;
		count++;
		if (is_rmr) {
			mpi3mr_add_sg_single(&rmgmt_req->response_sgl,
			    sgl_flags_last, dma_buff->kern_buf_len,
			    dma_buff->kern_buf_dma);
			dma_buff++;
			count++;
		} else
			mpi3mr_build_zero_len_sge(
			    &rmgmt_req->response_sgl);
	}
	if (!num_datasges) {
		mpi3mr_build_zero_len_sge(sgl);
		return;
	}
	for (; count < bufcnt; count++, dma_buff++) {
		if (dma_buff->data_dir == DMA_BIDIRECTIONAL)
			continue;
		if (num_datasges == 1 || !is_rmc)
			mpi3mr_add_sg_single(sgl, sgl_flags_last,
			    dma_buff->kern_buf_len, dma_buff->kern_buf_dma);
		else
			mpi3mr_add_sg_single(sgl, sgl_flags,
			    dma_buff->kern_buf_len, dma_buff->kern_buf_dma);
		sgl += sizeof(struct mpi3_sge_common);
		num_datasges--;
	}
}

/**
 * mpi3mr_get_nvme_data_fmt - returns the NVMe data format
 * @nvme_encap_request: NVMe encapsulated MPI request
 *
 * This function returns the type of the data format specified
 * in user provided NVMe command in NVMe encapsulated request.
 *
 * Return: Data format of the NVMe command (PRP/SGL etc)
 */
static unsigned int mpi3mr_get_nvme_data_fmt(
	struct mpi3_nvme_encapsulated_request *nvme_encap_request)
{
	u8 format = 0;

	format = ((nvme_encap_request->command[0] & 0xc000) >> 14);
	return format;

}

/**
 * mpi3mr_build_nvme_sgl - SGL constructor for NVME
 *				   encapsulated request
 * @mrioc: Adapter instance reference
 * @nvme_encap_request: NVMe encapsulated MPI request
 * @dma_buffers: DMA address of the buffers to be placed in sgl
 * @bufcnt: Number of DMA buffers
 *
 * This function places the DMA address of the given buffers in
 * proper format as SGEs in the given NVMe encapsulated request.
 *
 * Return: 0 on success, -1 on failure
 */
static int mpi3mr_build_nvme_sgl(struct mpi3mr_ioc *mrioc,
	struct mpi3_nvme_encapsulated_request *nvme_encap_request,
	struct mpi3mr_buf_map *dma_buffers, u8 bufcnt)
{
	struct mpi3mr_nvme_pt_sge *nvme_sgl;
	u64 sgl_ptr;
	u8 count;
	size_t length = 0;
	struct mpi3mr_buf_map *dma_buff = dma_buffers;
	u64 sgemod_mask = ((u64)((mrioc->facts.sge_mod_mask) <<
			    mrioc->facts.sge_mod_shift) << 32);
	u64 sgemod_val = ((u64)(mrioc->facts.sge_mod_value) <<
			  mrioc->facts.sge_mod_shift) << 32;

	/*
	 * Not all commands require a data transfer. If no data, just return
	 * without constructing any sgl.
	 */
	for (count = 0; count < bufcnt; count++, dma_buff++) {
		if ((dma_buff->data_dir == DMA_TO_DEVICE) ||
		    (dma_buff->data_dir == DMA_FROM_DEVICE)) {
			sgl_ptr = (u64)dma_buff->kern_buf_dma;
			length = dma_buff->kern_buf_len;
			break;
		}
	}
	if (!length)
		return 0;

	if (sgl_ptr & sgemod_mask) {
		dprint_ioctl_err(mrioc,
		    "%s: SGL address collides with SGE modifier\n",
		    __func__);
		return -1;
	}

	sgl_ptr &= ~sgemod_mask;
	sgl_ptr |= sgemod_val;
	nvme_sgl = (struct mpi3mr_nvme_pt_sge *)
	    ((u8 *)(nvme_encap_request->command) + MPI3MR_NVME_CMD_SGL_OFFSET);
	memset(nvme_sgl, 0, sizeof(struct mpi3mr_nvme_pt_sge));
	nvme_sgl->base_addr = sgl_ptr;
	nvme_sgl->length = length;
	return 0;
}

/**
 * mpi3mr_build_nvme_prp - PRP constructor for NVME
 *			       encapsulated request
 * @mrioc: Adapter instance reference
 * @nvme_encap_request: NVMe encapsulated MPI request
 * @dma_buffers: DMA address of the buffers to be placed in SGL
 * @bufcnt: Number of DMA buffers
 *
 * This function places the DMA address of the given buffers in
 * proper format as PRP entries in the given NVMe encapsulated
 * request.
 *
 * Return: 0 on success, -1 on failure
 */
static int mpi3mr_build_nvme_prp(struct mpi3mr_ioc *mrioc,
    struct mpi3_nvme_encapsulated_request *nvme_encap_request,
    struct mpi3mr_buf_map *dma_buffers, u8 bufcnt)
{
	int prp_size = MPI3MR_NVME_PRP_SIZE;
	__le64 *prp_entry, *prp1_entry, *prp2_entry;
	__le64 *prp_page;
	dma_addr_t prp_entry_dma, prp_page_dma, dma_addr;
	u32 offset, entry_len, dev_pgsz;
	u32 page_mask_result, page_mask;
	size_t length = 0;
	u8 count;
	struct mpi3mr_buf_map *dma_buff = dma_buffers;
	u64 sgemod_mask = ((u64)((mrioc->facts.sge_mod_mask) <<
			    mrioc->facts.sge_mod_shift) << 32);
	u64 sgemod_val = ((u64)(mrioc->facts.sge_mod_value) <<
			  mrioc->facts.sge_mod_shift) << 32;
	u16 dev_handle = nvme_encap_request->dev_handle;
	struct mpi3mr_tgt_dev *tgtdev;

	tgtdev = mpi3mr_get_tgtdev_by_handle(mrioc, dev_handle);
	if (!tgtdev) {
		dprint_ioctl_err(mrioc, "%s: invalid device handle 0x%04x\n",
			__func__, dev_handle);
		return -1;
	}
	if (tgtdev->dev_spec.pcie_inf.pgsz == 0) {
		dprint_ioctl_err(mrioc,
		    "%s: NVMe device page size is zero for handle 0x%04x\n",
		    __func__, dev_handle);
		mpi3mr_tgtdev_put(tgtdev);
		return -1;
	}
	dev_pgsz = 1 << (tgtdev->dev_spec.pcie_inf.pgsz);
	mpi3mr_tgtdev_put(tgtdev);

	/*
	 * Not all commands require a data transfer. If no data, just return
	 * without constructing any PRP.
	 */
	for (count = 0; count < bufcnt; count++, dma_buff++) {
		if ((dma_buff->data_dir == DMA_TO_DEVICE) ||
		    (dma_buff->data_dir == DMA_FROM_DEVICE)) {
			dma_addr = dma_buff->kern_buf_dma;
			length = dma_buff->kern_buf_len;
			break;
		}
	}
	if (!length)
		return 0;

	mrioc->prp_sz = 0;
	mrioc->prp_list_virt = dma_zalloc_coherent(&mrioc->pdev->dev,
	    dev_pgsz, &mrioc->prp_list_dma, GFP_KERNEL);

	if (!mrioc->prp_list_virt)
		return -1;
	mrioc->prp_sz = dev_pgsz;

	/*
	 * Set pointers to PRP1 and PRP2, which are in the NVMe command.
	 * PRP1 is located at a 24 byte offset from the start of the NVMe
	 * command.  Then set the current PRP entry pointer to PRP1.
	 */
	prp1_entry = (__le64 *)((u8 *)(nvme_encap_request->command) +
	    MPI3MR_NVME_CMD_PRP1_OFFSET);
	prp2_entry = (__le64 *)((u8 *)(nvme_encap_request->command) +
	    MPI3MR_NVME_CMD_PRP2_OFFSET);
	prp_entry = prp1_entry;
	/*
	 * For the PRP entries, use the specially allocated buffer of
	 * contiguous memory.
	 */
	prp_page = (__le64 *)mrioc->prp_list_virt;
	prp_page_dma = mrioc->prp_list_dma;

	/*
	 * Check if we are within 1 entry of a page boundary we don't
	 * want our first entry to be a PRP List entry.
	 */
	page_mask = dev_pgsz - 1;
	page_mask_result = (uintptr_t)((u8 *)prp_page + prp_size) & page_mask;
	if (!page_mask_result) {
		dprint_ioctl_err(mrioc, "%s: PRP page is not page aligned\n",
		    __func__);
		goto err_out;
	}

	/*
	 * Set PRP physical pointer, which initially points to the current PRP
	 * DMA memory page.
	 */
	prp_entry_dma = prp_page_dma;


	/* Loop while the length is not zero. */
	while (length) {
		page_mask_result = (prp_entry_dma + prp_size) & page_mask;
		if (!page_mask_result && (length >  dev_pgsz)) {
			dprint_ioctl_err(mrioc,
			    "%s: single PRP page is not sufficient\n",
			    __func__);
			goto err_out;
		}

		/* Need to handle if entry will be part of a page. */
		offset = dma_addr & page_mask;
		entry_len = dev_pgsz - offset;

		if (prp_entry == prp1_entry) {
			/*
			 * Must fill in the first PRP pointer (PRP1) before
			 * moving on.
			 */
			*prp1_entry = cpu_to_le64(dma_addr);
			if (*prp1_entry & sgemod_mask) {
				dprint_ioctl_err(mrioc,
				    "%s: PRP1 address collides with SGE modifier\n",
				    __func__);
				goto err_out;
			}
			*prp1_entry &= ~sgemod_mask;
			*prp1_entry |= sgemod_val;

			/*
			 * Now point to the second PRP entry within the
			 * command (PRP2).
			 */
			prp_entry = prp2_entry;
		} else if (prp_entry == prp2_entry) {
			/*
			 * Should the PRP2 entry be a PRP List pointer or just
			 * a regular PRP pointer?  If there is more than one
			 * more page of data, must use a PRP List pointer.
			 */
			if (length > dev_pgsz) {
				/*
				 * PRP2 will contain a PRP List pointer because
				 * more PRP's are needed with this command. The
				 * list will start at the beginning of the
				 * contiguous buffer.
				 */
				*prp2_entry = cpu_to_le64(prp_entry_dma);
				if (*prp2_entry & sgemod_mask) {
					dprint_ioctl_err(mrioc,
					    "%s: PRP list address collides with SGE modifier\n",
					    __func__);
					goto err_out;
				}
				*prp2_entry &= ~sgemod_mask;
				*prp2_entry |= sgemod_val;

				/*
				 * The next PRP Entry will be the start of the
				 * first PRP List.
				 */
				prp_entry = prp_page;
				continue;
			} else {
				/*
				 * After this, the PRP Entries are complete.
				 * This command uses 2 PRP's and no PRP list.
				 */
				*prp2_entry = cpu_to_le64(dma_addr);
				if (*prp2_entry & sgemod_mask) {
					dprint_ioctl_err(mrioc,
					    "%s: PRP2 collides with SGE modifier\n",
					    __func__);
					goto err_out;
				}
				*prp2_entry &= ~sgemod_mask;
				*prp2_entry |= sgemod_val;
			}
		} else {
			/*
			 * Put entry in list and bump the addresses.
			 *
			 * After PRP1 and PRP2 are filled in, this will fill in
			 * all remaining PRP entries in a PRP List, one per
			 * each time through the loop.
			 */
			*prp_entry = cpu_to_le64(dma_addr);
			if (*prp1_entry & sgemod_mask) {
				dprint_ioctl_err(mrioc,
				    "%s: PRP address collides with SGE modifier\n",
				    __func__);
				goto err_out;
			}
			*prp_entry &= ~sgemod_mask;
			*prp_entry |= sgemod_val;
			prp_entry++;
			prp_entry_dma++;
		}

		/*
		 * Bump the phys address of the command's data buffer by the
		 * entry_len.
		 */
		dma_addr += entry_len;

		/* decrement length accounting for last partial page. */
		if (entry_len > length)
			length = 0;
		else
			length -= entry_len;
	}
	return 0;
err_out:
	if (mrioc->prp_list_virt) {
		dma_free_coherent(&mrioc->pdev->dev, mrioc->prp_sz,
		    mrioc->prp_list_virt, mrioc->prp_list_dma);
		mrioc->prp_list_virt = NULL;
	}
	return -1;
}

/**
 * mpi3mr_ioctl_process_mpt_cmds - MPI Pass through IOCTL handler
 * @file: File pointer
 * @arg: User data payload buffer for the IOCTL
 *
 * This function is the top level handler for MPI Pass through
 * IOCTL, this does basic validation of the input data buffers,
 * identifies the given buffer types and MPI command, allocates
 * DMAable memory for user given buffers, construstcs SGL
 * properly and passes the command to the firmware.
 *
 * Once the MPI command is completed the driver copies the data
 * if any and reply, sense information to user provided buffers.
 * If the command is timed out then issues controller reset
 * prior to returning.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_ioctl_process_mpt_cmds(struct file *file, void __user *arg)
{
	long rval = -EINVAL;
	struct mpi3mr_ioc *mrioc = NULL;
	struct mpi3mr_ioctl_mptcmd karg;
	u8 *mpi_req = NULL;
	struct mpi3mr_ioctl_buf_entry_list *buffer_list = NULL;
	struct mpi3mr_buf_entry *buf_entries = NULL;
	struct mpi3mr_buf_map *dma_buffers = NULL, *dma_buff;
	u8 count, bufcnt, is_rmcb = 0, is_rmrb = 0, din_cnt = 0, dout_cnt = 0;
	u8 invalid_be = 0, erb_offset = 0xFF, mpirep_offset = 0xFF;
	u8 sg_entries = 0;
	u32 tmplen;
	struct mpi3_request_header *mpi_header = NULL;
	struct mpi3_status_reply_descriptor *status_desc;
	struct mpi3_scsi_task_mgmt_request *tm_req;
	u8 *sense_buff_k = NULL;
	u32 erbsz = MPI3MR_SENSE_BUF_SZ;
	u8 block_io = 0;
	u8 nvme_fmt = 0;
	u16 dev_handle;
	u8 resp_code = 0;
	struct mpi3mr_tgt_dev *tgtdev;
	struct mpi3mr_stgt_priv_data *stgt_priv = NULL;
	struct mpi3mr_ioctl_in_ioctl_reply_buf *ioctl_reply_buf = NULL;
	u32 din_size = 0, dout_size = 0;

	if (copy_from_user(&karg, arg, sizeof(karg)))
		return -EFAULT;
	mpi3mr_app_verify_adapter(karg.mrioc_id, &mrioc);
	if (!mrioc)
		return -ENODEV;

	if (karg.timeout < MPI3MR_IOCTL_DEFAULT_TIMEOUT)
		karg.timeout = MPI3MR_IOCTL_DEFAULT_TIMEOUT;

	if (!karg.buf_entry_list_size || !karg.mpi_msg_size) {
		dprint_ioctl_err(mrioc, "%s: invalid size parameters passed\n",
		    __func__);
		return -EINVAL;
	}
	if ((karg.mpi_msg_size * 4) > MPI3MR_ADMIN_REQ_FRAME_SZ) {
		dprint_ioctl_err(mrioc, "%s:%d: invalid message size passed\n",
		    __func__, __LINE__);
		return -EINVAL;
	}

	mpi_req = kzalloc(MPI3MR_ADMIN_REQ_FRAME_SZ, GFP_KERNEL);
	if (!mpi_req)
		return -ENOMEM;
	mpi_header = (struct mpi3_request_header *)mpi_req;

	if (copy_from_user(mpi_req, karg.mpi_msg_buf,
	    (karg.mpi_msg_size * 4))) {
		dprint_ioctl_err(mrioc, "%s:%d: copy from user failed\n",
		    __func__, __LINE__);
		rval = -EFAULT;
		goto out;
	}

	buffer_list = kzalloc(karg.buf_entry_list_size, GFP_KERNEL);
	if (!buffer_list) {
		rval = -ENOMEM;
		goto out;
	}
	if (copy_from_user(buffer_list, karg.buf_entry_list,
	    karg.buf_entry_list_size)) {
		dprint_ioctl_err(mrioc, "%s:%d: copy from user failed\n",
		    __func__, __LINE__);
		rval = -EFAULT;
		goto out;
	}

	if (!buffer_list->num_of_entries) {
		dprint_ioctl_err(mrioc, "%s: no buffer entries passed\n",
		    __func__);
		rval = -EINVAL;
		goto out;
	}
	bufcnt = buffer_list->num_of_entries;
	dma_buffers = kzalloc((sizeof(*dma_buffers) * bufcnt), GFP_KERNEL);
	if (!dma_buffers) {
		rval = -ENOMEM;
		goto out;
	}
	buf_entries = buffer_list->buf_entry;
	dma_buff = dma_buffers;
	for (count = 0; count < bufcnt; count++, buf_entries++, dma_buff++) {
		dma_buff->user_buf = buf_entries->buffer;
		dma_buff->user_buf_len = buf_entries->buf_len;

		switch (buf_entries->buf_type) {
		case MPI3MR_IOCTL_BUFTYPE_RAIDMGMT_CMD:
			is_rmcb = 1;
			if (count != 0)
				invalid_be = 1;
			dma_buff->data_dir = DMA_FROM_DEVICE;
			break;
		case MPI3MR_IOCTL_BUFTYPE_RAIDMGMT_RESP:
			is_rmrb = 1;
			if (count != 1 || !is_rmcb)
				invalid_be = 1;
			dma_buff->data_dir = DMA_TO_DEVICE;
			break;
		case MPI3MR_IOCTL_BUFTYPE_DATA_IN:
			din_cnt++;
			din_size += dma_buff->user_buf_len;
			if ((din_cnt > 1) && !is_rmcb)
				invalid_be = 1;
			dma_buff->data_dir = DMA_TO_DEVICE;
			break;
		case MPI3MR_IOCTL_BUFTYPE_DATA_OUT:
			dout_cnt++;
			dout_size += dma_buff->user_buf_len;
			if ((dout_cnt > 1) && !is_rmcb)
				invalid_be = 1;
			dma_buff->data_dir = DMA_FROM_DEVICE;
			break;
		case MPI3MR_IOCTL_BUFTYPE_MPI_REPLY:
			mpirep_offset = count;
			dma_buff->data_dir = DMA_BIDIRECTIONAL;
			break;
		case MPI3MR_IOCTL_BUFTYPE_ERR_RESPONSE:
			erb_offset = count;
			dma_buff->data_dir = DMA_BIDIRECTIONAL;
			break;
		default:
			invalid_be = 1;
			break;
		}
		if (invalid_be)
			break;
	}
	if (invalid_be) {
		dprint_ioctl_err(mrioc, "%s: invalid buffer entries passed\n",
			__func__);
		rval = -EINVAL;
		goto out;
	}

	if (!is_rmcb && (dout_cnt || din_cnt)) {
		sg_entries = dout_cnt + din_cnt;
		if (((karg.mpi_msg_size * 4) + (sg_entries *
		      sizeof(struct mpi3_sge_common))) > MPI3MR_ADMIN_REQ_FRAME_SZ) {
			dprint_ioctl_err(mrioc,
			    "%s:%d: invalid message size passed\n",
			    __func__, __LINE__);
			rval = -EINVAL;
			goto out;
		}
	}
	if (din_size > MPI3MR_MAX_IOCTL_XFER_SIZE) {
		dprint_ioctl_err(mrioc,
		    "%s:%d: invalid data transfer size passed for function 0x%x din_size=%d\n",
		    __func__, __LINE__, mpi_header->function, din_size);
		rval = -EINVAL;
		goto out;
	}
	if (dout_size > MPI3MR_MAX_IOCTL_XFER_SIZE) {
		dprint_ioctl_err(mrioc,
		    "%s:%d: invalid data transfer size passed for function 0x%x dout_size = %d\n",
		    __func__, __LINE__, mpi_header->function, dout_size);
		rval = -EINVAL;
		goto out;
	}

	dma_buff = dma_buffers;
	for (count = 0; count < bufcnt; count++, dma_buff++) {
		dma_buff->kern_buf_len = dma_buff->user_buf_len;
		if (is_rmcb && !count)
			dma_buff->kern_buf_len += ((dout_cnt + din_cnt) *
			    sizeof(struct mpi3_sge_common));
		if ((count == mpirep_offset) || (count == erb_offset)) {
			dma_buff->kern_buf_len = 0;
			continue;
		}
		if (!dma_buff->kern_buf_len)
			continue;

		dma_buff->kern_buf = dma_zalloc_coherent(&mrioc->pdev->dev,
		    dma_buff->kern_buf_len, &dma_buff->kern_buf_dma,
		    GFP_KERNEL);
		if (!dma_buff->kern_buf) {
			rval = -ENOMEM;
			goto out;
		}
		if (dma_buff->data_dir == DMA_FROM_DEVICE) {
			tmplen = min(dma_buff->kern_buf_len,
			    dma_buff->user_buf_len);
			if (copy_from_user(dma_buff->kern_buf,
			    dma_buff->user_buf, tmplen)) {
				dprint_ioctl_err(mrioc,
				    "%s:%d: copy from user failed\n",
				    __func__, __LINE__);
				rval = -EFAULT;
				goto out;
			}
		}
	}
	if (erb_offset != 0xFF) {
		sense_buff_k = kzalloc(erbsz, GFP_KERNEL);
		if (!sense_buff_k) {
			rval = -ENOMEM;
			goto out;
		}
	}

	if (file->f_flags & O_NONBLOCK) {
		if (!mutex_trylock(&mrioc->ioctl_cmds.mutex)) {
			rval = -EAGAIN;
			goto out;
		}
	} else if (mutex_lock_interruptible(&mrioc->ioctl_cmds.mutex)) {
		rval = -ERESTARTSYS;
		goto out;
	}
	if (mrioc->ioctl_cmds.state & MPI3MR_CMD_PENDING) {
		rval = -EAGAIN;
		dprint_ioctl_err(mrioc, "%s: command is in use\n", __func__);
		mutex_unlock(&mrioc->ioctl_cmds.mutex);
		goto out;
	}
	if (mrioc->unrecoverable) {
		dprint_ioctl_err(mrioc, "%s: unrecoverable controller\n",
		    __func__);
		rval = -EFAULT;
		mutex_unlock(&mrioc->ioctl_cmds.mutex);
		goto out;
	}
	if (mrioc->reset_in_progress) {
		dprint_ioctl_err(mrioc, "%s: reset in progress\n", __func__);
		rval = -EAGAIN;
		mutex_unlock(&mrioc->ioctl_cmds.mutex);
		goto out;
	}
	if (mrioc->block_ioctls) {
		dprint_ioctl_err(mrioc, "%s: ioctls are blocked\n", __func__);
		rval = -EAGAIN;
		mutex_unlock(&mrioc->ioctl_cmds.mutex);
		goto out;
	}

	if (mpi_header->function != MPI3_FUNCTION_NVME_ENCAPSULATED) {
		mpi3mr_ioctl_build_sgl(mpi_req, (karg.mpi_msg_size * 4),
		    dma_buffers, bufcnt, is_rmcb, is_rmrb,
		    (dout_cnt + din_cnt));
	}

	if (mpi_header->function == MPI3_FUNCTION_NVME_ENCAPSULATED) {
		nvme_fmt = mpi3mr_get_nvme_data_fmt(
			(struct mpi3_nvme_encapsulated_request *)mpi_req);
		if (nvme_fmt == MPI3MR_NVME_DATA_FORMAT_PRP) {
			if (mpi3mr_build_nvme_prp(mrioc,
			    (struct mpi3_nvme_encapsulated_request *)mpi_req,
			    dma_buffers, bufcnt)) {
				rval = -ENOMEM;
				mutex_unlock(&mrioc->ioctl_cmds.mutex);
				goto out;
			}
		} else if (nvme_fmt == MPI3MR_NVME_DATA_FORMAT_SGL1 ||
			nvme_fmt == MPI3MR_NVME_DATA_FORMAT_SGL2) {
			if (mpi3mr_build_nvme_sgl(mrioc,
			    (struct mpi3_nvme_encapsulated_request *)mpi_req,
			    dma_buffers, bufcnt)) {
				rval = -EINVAL;
				mutex_unlock(&mrioc->ioctl_cmds.mutex);
				goto out;
			}
		} else {
			dprint_ioctl_err(mrioc,
			    "%s:invalid NVMe command format\n", __func__);
			rval = -EINVAL;
			mutex_unlock(&mrioc->ioctl_cmds.mutex);
			goto out;
		}
	}
	if (mpi_header->function == MPI3_FUNCTION_SCSI_TASK_MGMT) {
		tm_req = (struct mpi3_scsi_task_mgmt_request *)mpi_req;
		if (tm_req->task_type !=
		    MPI3_SCSITASKMGMT_TASKTYPE_ABORT_TASK) {
			dev_handle = tm_req->dev_handle;
			block_io = 1;
		}
	}
	if (block_io) {
		tgtdev = mpi3mr_get_tgtdev_by_handle(mrioc, dev_handle);
		if (tgtdev && tgtdev->starget && tgtdev->starget->hostdata) {
			stgt_priv = (struct mpi3mr_stgt_priv_data *)
			    tgtdev->starget->hostdata;
			atomic_inc(&stgt_priv->block_io);
			mpi3mr_tgtdev_put(tgtdev);
		}
	}

	mrioc->ioctl_cmds.state = MPI3MR_CMD_PENDING;
	mrioc->ioctl_cmds.is_waiting = 1;
	mrioc->ioctl_cmds.callback = NULL;
	mrioc->ioctl_cmds.is_sense = 0;
	mrioc->ioctl_cmds.sensebuf = sense_buff_k;
	memset(mrioc->ioctl_cmds.reply, 0, mrioc->reply_sz);
	mpi_header->host_tag = cpu_to_le16(MPI3MR_HOSTTAG_IOCTLCMDS);
	if (mrioc->logging_level & MPI3_DEBUG_IOCTL_INFO) {
		dprint_ioctl_info(mrioc,
		    "%s: posting ioctl request to the controller\n", __func__);
		dprint_dump(mpi_req, MPI3MR_ADMIN_REQ_FRAME_SZ,
		    "ioctl_mpi3_req");
		if (mpi_header->function == MPI3_FUNCTION_MGMT_PASSTHROUGH) {
			dma_buff = &dma_buffers[0];
			dprint_dump(dma_buff->kern_buf,
			    dma_buff->kern_buf_len, "mpi3_mgmt_req");
		}
	}

	init_completion(&mrioc->ioctl_cmds.done);
	rval = mpi3mr_admin_request_post(mrioc, mpi_req,
	    MPI3MR_ADMIN_REQ_FRAME_SZ, 0);

	if (rval) {
		mrioc->ioctl_cmds.is_waiting = 0;
		dprint_ioctl_err(mrioc,
		    "%s: posting ioctl request is failed\n", __func__);
		rval = -EAGAIN;
		goto out_unlock;
	}
	wait_for_completion_timeout(&mrioc->ioctl_cmds.done,
	    (karg.timeout * HZ));
	if (block_io && stgt_priv)
		atomic_dec(&stgt_priv->block_io);
	if (!(mrioc->ioctl_cmds.state & MPI3MR_CMD_COMPLETE)) {
		mrioc->ioctl_cmds.is_waiting = 0;
		rval = -EAGAIN;
		if (mrioc->ioctl_cmds.state & MPI3MR_CMD_RESET)
			goto out_unlock;
		dprint_ioctl_err(mrioc,
		    "%s: ioctl request timedout after %d seconds\n", __func__,
		    karg.timeout);
		if (mrioc->logging_level & MPI3_DEBUG_IOCTL_ERROR) {
			dprint_dump(mpi_req, MPI3MR_ADMIN_REQ_FRAME_SZ,
			    "ioctl_mpi3_req");
			if (mpi_header->function ==
			    MPI3_FUNCTION_MGMT_PASSTHROUGH) {
				dma_buff = &dma_buffers[0];
				dprint_dump(dma_buff->kern_buf,
				    dma_buff->kern_buf_len, "mpi3_mgmt_req");
			}
		}
		if ((mpi_header->function == MPI3_FUNCTION_NVME_ENCAPSULATED) ||
		    (mpi_header->function == MPI3_FUNCTION_SCSI_IO))
			mpi3mr_issue_tm(mrioc,
			    MPI3_SCSITASKMGMT_TASKTYPE_TARGET_RESET,
			    mpi_header->function_dependent, 0,
			    MPI3MR_HOSTTAG_BLK_TMS, MPI3MR_RESETTM_TIMEOUT,
			    &mrioc->host_tm_cmds, &resp_code, NULL);
		if (!(mrioc->ioctl_cmds.state & MPI3MR_CMD_COMPLETE) &&
		    !(mrioc->ioctl_cmds.state & MPI3MR_CMD_RESET))
			mpi3mr_soft_reset_handler(mrioc,
			    MPI3MR_RESET_FROM_IOCTL_TIMEOUT, 1);
		goto out_unlock;
	}
	dprint_ioctl_info(mrioc, "%s: ioctl request is completed\n", __func__);

	if (mrioc->prp_list_virt) {
		dma_free_coherent(&mrioc->pdev->dev, mrioc->prp_sz,
		    mrioc->prp_list_virt, mrioc->prp_list_dma);
		mrioc->prp_list_virt = NULL;
	}

	if ((mrioc->ioctl_cmds.ioc_status & MPI3_IOCSTATUS_STATUS_MASK)
	     != MPI3_IOCSTATUS_SUCCESS) {
		dprint_ioctl_info(mrioc,
		    "%s: command failed, ioc_status(0x%04x) log_info(0x%08x)\n",
		    __func__,
		    (mrioc->ioctl_cmds.ioc_status & MPI3_IOCSTATUS_STATUS_MASK),
		    mrioc->ioctl_cmds.ioc_loginfo);
	}

	if ((mpirep_offset != 0xFF) &&
	    dma_buffers[mpirep_offset].user_buf_len) {
		dma_buff = &dma_buffers[mpirep_offset];
		dma_buff->kern_buf_len = (sizeof(*ioctl_reply_buf) - 1 +
					   mrioc->reply_sz);
		ioctl_reply_buf = kzalloc(dma_buff->kern_buf_len, GFP_KERNEL);

		if (!ioctl_reply_buf) {
			rval = -ENOMEM;
			goto out_unlock;
		}
		if (mrioc->ioctl_cmds.state & MPI3MR_CMD_REPLY_VALID) {
			ioctl_reply_buf->mpi_reply_type =
				MPI3MR_IOCTL_MPI_REPLY_BUFTYPE_ADDRESS;
			memcpy(ioctl_reply_buf->ioctl_reply_buf,
			    mrioc->ioctl_cmds.reply, mrioc->reply_sz);
		} else {
			ioctl_reply_buf->mpi_reply_type =
				MPI3MR_IOCTL_MPI_REPLY_BUFTYPE_STATUS;
			status_desc = (struct mpi3_status_reply_descriptor *)
			    ioctl_reply_buf->ioctl_reply_buf;
			status_desc->ioc_status = mrioc->ioctl_cmds.ioc_status;
			status_desc->ioc_log_info = mrioc->ioctl_cmds.ioc_loginfo;
		}
		tmplen = min(dma_buff->kern_buf_len, dma_buff->user_buf_len);
		if (copy_to_user(dma_buff->user_buf, ioctl_reply_buf, tmplen)) {
			dprint_ioctl_err(mrioc, "%s:%d: copy to user failed\n",
			    __func__, __LINE__);
			rval = -EFAULT;
			goto out_unlock;
		}
	}

	if (erb_offset != 0xFF && mrioc->ioctl_cmds.sensebuf &&
	    mrioc->ioctl_cmds.is_sense) {
		dma_buff = &dma_buffers[erb_offset];
		tmplen = min(erbsz, dma_buff->user_buf_len);
		if (copy_to_user(dma_buff->user_buf, sense_buff_k, tmplen)) {
			dprint_ioctl_err(mrioc, "%s:%d: copy to user failed\n",
			    __func__, __LINE__);
			rval = -EFAULT;
			goto out_unlock;
		}
	}

	dma_buff = dma_buffers;
	for (count = 0; count < bufcnt; count++, dma_buff++) {
		if (dma_buff->data_dir == DMA_TO_DEVICE) {
			tmplen = min(dma_buff->kern_buf_len,
			    dma_buff->user_buf_len);
			if (copy_to_user(dma_buff->user_buf,
			    dma_buff->kern_buf, tmplen)) {
				dprint_ioctl_err(mrioc,
				    "%s:%d: copy to user failed\n",
				    __func__, __LINE__);
				rval = -EFAULT;
				goto out;
			}
		}
	}

out_unlock:
	mrioc->ioctl_cmds.is_sense = 0;
	mrioc->ioctl_cmds.sensebuf = NULL;
	mrioc->ioctl_cmds.state = MPI3MR_CMD_NOTUSED;
	mutex_unlock(&mrioc->ioctl_cmds.mutex);
out:
	kfree(sense_buff_k);
	kfree(buffer_list);
	kfree(mpi_req);
	if (dma_buffers) {
		dma_buff = dma_buffers;
		for (count = 0; count < bufcnt; count++, dma_buff++) {
			if (dma_buff->kern_buf && dma_buff->kern_buf_dma)
				dma_free_coherent(&mrioc->pdev->dev,
				    dma_buff->kern_buf_len,
				    dma_buff->kern_buf,
				    dma_buff->kern_buf_dma);
		}
		kfree(dma_buffers);
	}
	kfree(ioctl_reply_buf);
	return rval;
}

/**
 * mpi3mr_app_ioctl - Main IOCTL handler
 * @file: File pointer
 * @cmd: IOCTL command
 * @arg: User data payload buffer for the IOCTL
 *
 * This is main IOCTL handler which checks the command type and
 * executes proper sub handler specific for the command.
 *
 * Return: 0 on success and proper error codes on failure
 */
static long mpi3mr_app_ioctl(struct file *file, unsigned int cmd,
    unsigned long arg)
{
	long rval = -EINVAL;

	switch (cmd) {
	case MPI3MRDRVCMD:
		if (_IOC_SIZE(cmd) == sizeof(struct mpi3mr_ioctl_drv_cmd))
			rval = mpi3mr_app_process_drv_cmds(file,
			    (void __user *)arg);
		break;
	case MPI3MRMPTCMD:
		if (_IOC_SIZE(cmd) == sizeof(struct mpi3mr_ioctl_mptcmd))
			rval = mpi3mr_ioctl_process_mpt_cmds(file,
			    (void __user *)arg);
		break;
	default:
		pr_err("%s: unsupported ioctl cmd (0x%08x)\n",
		    MPI3MR_DRIVER_NAME, cmd);
		break;
	}
	return rval;
}

/**
 * mpi3mr_app_send_aen - Notify applications about an AEN
 * @mrioc: Adapter instance reference
 *
 * Sends async signal SIGIO to indicate there is an async event
 * from the firmware to the event monitoring applications.
 *
 * Return:Nothing
 */
void mpi3mr_app_send_aen(struct mpi3mr_ioc *mrioc)
{
	dprint_ioctl_info(mrioc, "%s: invoked\n", __func__);
	if (mpi3mr_app_async_queue) {
		dprint_ioctl_info(mrioc, "%s: sending signal\n", __func__);
		kill_fasync(&mpi3mr_app_async_queue, SIGIO, POLL_IN);
	}
}

/**
 * mpi3mr_app_save_logdata - Save Log Data events
 * @mrioc: Adapter instance reference
 * @event_data: event data associated with log data event
 * @event_data_size: event data size to copy
 *
 * If log data event caching is enabled by the applicatiobns,
 * then this function saves the log data in the circular queue
 * and Sends async signal SIGIO to indicate there is an async
 * event from the firmware to the event monitoring applications.
 *
 * Return:Nothing
 */
void mpi3mr_app_save_logdata(struct mpi3mr_ioc *mrioc, char *event_data,
    u16 event_data_size)
{
	u32 index = mrioc->logdata_buf_idx, sz;
	struct mpi3mr_logdata_entry *entry;

	if (!(mrioc->logdata_buf))
		return;

	entry = (struct mpi3mr_logdata_entry *)
		(mrioc->logdata_buf + (index * mrioc->logdata_entry_sz));
	entry->valid_entry = 1;
	sz = min(mrioc->logdata_entry_sz, event_data_size);
	memcpy(entry->data, event_data, sz);
	mrioc->logdata_buf_idx =
		((++index) % MPI3MR_IOCTL_LOGDATA_MAX_ENTRIES);
	mpi3mr_app_send_aen(mrioc);
}

/**
 * mpi3mr_app_fasync - fasync callback
 * @fd: File descriptor
 * @filep: File pointer
 * @mode: Mode
 *
 * Return: fasync_helper() returned value
 */
static int mpi3mr_app_fasync(int fd, struct file *filep, int mode)
{
	return fasync_helper(fd, filep, mode, &mpi3mr_app_async_queue);
}

static const struct file_operations mpi3mr_app_fops = {
	.owner = THIS_MODULE,
	.unlocked_ioctl = mpi3mr_app_ioctl,
	.fasync = mpi3mr_app_fasync,
};

static struct miscdevice mpi3mr_app_dev = {
	.minor  = MPI3MR_MINOR,
	.name   = MPI3MR_DEV_NAME,
	.fops   = &mpi3mr_app_fops,
};

/**
 * mpi3mr_app_init - Character driver interface initializer
 *
 */
void mpi3mr_app_init(void)
{
	mpi3mr_app_async_queue = NULL;

	if (misc_register(&mpi3mr_app_dev) < 0)
		pr_err("%s can't register misc device [minor=%d]\n",
		    MPI3MR_DRIVER_NAME, MPI3MR_MINOR);
}

/**
 * mpi3mr_app_exit - Character driver interface cleanup function
 *
 */
void mpi3mr_app_exit(void)
{
	misc_deregister(&mpi3mr_app_dev);
}

/*
 * SCSI Host attributes under sysfs
 */
/**
 * version_fw_show - SysFS callback for firmware version read
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 *
 * Return: snprintf() return after copying firmware version
 */
static ssize_t
version_fw_show(struct device *dev, struct device_attribute *attr,
	char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct mpi3mr_ioc *mrioc = shost_priv(shost);
	struct mpi3mr_compimg_ver *fwver = &mrioc->facts.fw_ver;

	return snprintf(buf, PAGE_SIZE, "%d.%d.%d.%d.%05d-%05d\n",
	    fwver->gen_major, fwver->gen_minor, fwver->ph_major,
	    fwver->ph_minor, fwver->cust_id, fwver->build_num);
}
static DEVICE_ATTR_RO(version_fw);

/**
 * fw_queue_depth_show - SysFS callback for firmware max cmds
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 *
 * Return: snprintf() return after copying firmware max commands
 */
static ssize_t
fw_queue_depth_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct mpi3mr_ioc *mrioc = shost_priv(shost);

	return snprintf(buf, PAGE_SIZE, "%d\n", mrioc->facts.max_reqs);
}
static DEVICE_ATTR_RO(fw_queue_depth);

/**
 * op_req_q_count_show - SysFS callback for request queue count
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 *
 * Return: snprintf() return after copying request queue count
 */
static ssize_t
op_req_q_count_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct mpi3mr_ioc *mrioc = shost_priv(shost);

	return snprintf(buf, PAGE_SIZE, "%d\n", mrioc->num_op_req_q);
}
static DEVICE_ATTR_RO(op_req_q_count);

/**
 * reply_queue_count_show - SysFS callback for reply queue count
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 *
 * Return: snprintf() return after copying reply queue count
 */
static ssize_t
reply_queue_count_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct mpi3mr_ioc *mrioc = shost_priv(shost);

	return snprintf(buf, PAGE_SIZE, "%d\n", mrioc->num_op_reply_q);
}

static DEVICE_ATTR_RO(reply_queue_count);

/**
 * mpi3mr_app_logging_level_show - Show controller debug level
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 *
 * A sysfs 'read/write' shost attribute, to show the current
 * debug log level used by the driver for the specific
 * controller.
 *
 * Return: snprintf() return
 */
static ssize_t
mpi3mr_app_logging_level_show(struct device *dev,
	struct device_attribute *attr, char *buf)

{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct mpi3mr_ioc *mrioc = shost_priv(shost);
	return snprintf(buf, PAGE_SIZE, "%08xh\n", mrioc->logging_level);
}

/**
 * mpi3mr_app_logging_level_store- Change controller debug level
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 * @count: size of the buffer
 *
 * A sysfs 'read/write' shost attribute, to change the current
 * debug log level used by the driver for the specific
 * controller.
 *
 * Return: strlen() return
 */
static ssize_t
mpi3mr_app_logging_level_store(struct device *dev,
	struct device_attribute *attr,
	const char *buf, size_t count)
{
	struct Scsi_Host *shost = class_to_shost(dev);
	struct mpi3mr_ioc *mrioc = shost_priv(shost);
	int val = 0;

	if (sscanf(buf, "%x", &val) != 1)
		return -EINVAL;

	mrioc->logging_level = val;
	ioc_info(mrioc, "logging_level=%08xh\n", mrioc->logging_level);
	return strlen(buf);
}
static DEVICE_ATTR(logging_level, 0644,
	mpi3mr_app_logging_level_show,
	mpi3mr_app_logging_level_store);


/**
 * mpi3mr_app_complete_tm - SysFS TM completion callback
 * @mrioc: Adapter instance reference
 * @drv_cmd: Internal command tracker
 *
 * This is a call back handler for the TM requests issued to the
 * firmware through SysFS interface in non blocking mode. This
 * functions wakes up pending TM wait queue event when all TMs
 * issued are completed.
 *
 * Return: nothing
 */
static void mpi3mr_app_complete_tm(struct mpi3mr_ioc *mrioc,
    struct mpi3mr_drv_cmd *drv_cmd)
{
	struct mpi3_scsi_task_mgmt_reply *tm_reply = NULL;
	u16 cmd_idx = drv_cmd->host_tag - MPI3MR_HOSTTAG_SYSFS_TM_MIN;

	if (drv_cmd->state & MPI3MR_CMD_RESET)
		goto clear_drv_cmd;

	if (drv_cmd->state & MPI3MR_CMD_REPLY_VALID) {
		tm_reply = (struct mpi3_scsi_task_mgmt_reply *)drv_cmd->reply;
		ioc_info(mrioc,
		    "%s:TM[%d]: completed, handle(0x%04x), ioc_status(0x%04x), log_info(0x%08x), termination_count(%d), response_code(0x%02x)\n",
		    __func__, cmd_idx+1, drv_cmd->dev_handle,
		    drv_cmd->ioc_status, drv_cmd->ioc_loginfo,
		    le32_to_cpu(tm_reply->termination_count),
		    (le32_to_cpu(tm_reply->response_data) &
		     MPI3MR_RI_MASK_RESPCODE));
		mrioc->sysfs_tm_terminated_io_count +=
			le32_to_cpu(tm_reply->termination_count);
	}
clear_drv_cmd:
	atomic_dec(&mrioc->sysfs_tm_pending);
	if (!atomic_read(&mrioc->sysfs_tm_pending))
		wake_up(&mrioc->sysfs_pending_tm_wq);
	drv_cmd->state = MPI3MR_CMD_NOTUSED;
	drv_cmd->callback = NULL;
	drv_cmd->dev_handle = MPI3MR_INVALID_DEV_HANDLE;
	return;
}

/**
 * mpi3mr_app_issue_tm - sends Task Management request
 * @mrioc: Adapter instance reference
 * @tm_type: Task Management type
 * @handle: Firmware device handle
 * @lun: lun ID
 * @cmd_priv: SCSI command private data
 *
 * This function sends Task Management request to the firmware
 * through admin request queue.
 *
 * Return: 0 on success, -1 on failure
 */
static int mpi3mr_app_issue_tm(struct mpi3mr_ioc *mrioc, u8 tm_type,
	u16 handle, uint lun, struct scmd_priv *cmd_priv)
{
	struct mpi3_scsi_task_mgmt_request tm_req;
	int r = -1;
	struct op_req_qinfo *op_req_q = NULL;
	struct mpi3mr_drv_cmd *drv_cmd;

	if ((mrioc->unrecoverable) || (mrioc->reset_in_progress) ||
	    (mrioc->sysfs_tm_issued >= MPI3MR_NUM_SYSFS_TM)) {
		return r;
	}
	drv_cmd = &mrioc->sysfs_tm_cmds[mrioc->sysfs_tm_issued];

	if (drv_cmd->state & MPI3MR_CMD_PENDING)
		return r;

	memset(&tm_req, 0, sizeof(tm_req));
	drv_cmd->state = MPI3MR_CMD_PENDING;
	drv_cmd->is_waiting = 0;
	drv_cmd->callback = mpi3mr_app_complete_tm;
	drv_cmd->dev_handle = handle;
	tm_req.dev_handle = cpu_to_le16(handle);
	tm_req.host_tag = cpu_to_le16(drv_cmd->host_tag);
	tm_req.function = MPI3_FUNCTION_SCSI_TASK_MGMT;
	tm_req.task_type = tm_type;

	int_to_scsilun(lun, (struct scsi_lun *)tm_req.lun);

	if (cmd_priv) {
		op_req_q = &mrioc->req_qinfo[cmd_priv->req_q_idx];
		tm_req.task_host_tag = cpu_to_le16(cmd_priv->host_tag);
		tm_req.task_request_queue_id = cpu_to_le16(op_req_q->qid);
	}

	ioc_info(mrioc, "%s: TM[%d] type (0x%02x) issued for handle (0x%04x)\n",
		__func__, mrioc->sysfs_tm_issued + 1, tm_type, handle);
	atomic_inc(&mrioc->sysfs_tm_pending);
	r = mpi3mr_admin_request_post(mrioc, &tm_req, sizeof(tm_req), 1);
	if (r) {
		atomic_dec(&mrioc->sysfs_tm_pending);
		ioc_err(mrioc, "%s : posting TM[%d] failed\n", __func__,
		    mrioc->sysfs_tm_issued + 1);
	} else
		mrioc->sysfs_tm_issued++;

	return r;
}

/**
 * mpi3mr_app_issue_abort_task - sends Task Abort
 * @rq: Block I/O request
 * @data: Adapter instance
 * @reserved: Unused
 *
 * This function sends Abort Task Management request to the
 * firmware, this is iterator callback for every I/O.
 *
 * Return: 0 on success, -1 on failure
 */
static BLK_ITER_CALLBACK_RET_TYPE mpi3mr_app_issue_abort_task(
    struct request *rq, void *data, bool reserved)
{
	struct mpi3mr_ioc *mrioc = (struct mpi3mr_ioc *)data;
	struct scsi_cmnd *scmd = blk_mq_rq_to_pdu(rq);
	struct scmd_priv *priv = NULL;
	struct mpi3mr_stgt_priv_data *stgt_priv_data;
	struct mpi3mr_sdev_priv_data *sdev_priv_data;
	u16 dev_handle;

	if (scmd) {
		sdev_priv_data = scmd->device->hostdata;
		if (sdev_priv_data && sdev_priv_data->tgt_priv_data) {
			stgt_priv_data = sdev_priv_data->tgt_priv_data;
			dev_handle = stgt_priv_data->dev_handle;
			priv = scsi_cmd_priv(scmd);
		}
	}

	if (priv && priv->in_lld_scope &&
	    (mrioc->sysfs_tm_issued < MPI3MR_NUM_SYSFS_TM)) {
		mpi3mr_app_issue_tm(mrioc,
		    MPI3_SCSITASKMGMT_TASKTYPE_ABORT_TASK, dev_handle,
		    sdev_priv_data->lun_id, priv);
		BLK_ITER_CALLBACK_RET_VAL(true);
	}

	BLK_ITER_CALLBACK_RET_VAL(false);
}

/**
 * mpi3mr_app_tm_sysfs - sends TM of given type
 * @mrioc: Adapter instance reference
 * @tm_type: Task Management type
 *
 * This function checks TM type and issue appropriate number of
 * specific TM to the devices/IO requests under the scope of the
 * TM and the controller and waits the TM requests to complete.
 * If TM requests are not completed within predefined timeout
 * then issues controller reset
 *
 * Return: Nothing
 */
static void mpi3mr_app_tm_sysfs(struct mpi3mr_ioc *mrioc, u8 tm_type)
{
	struct mpi3mr_tgt_dev *tgtdev;
	struct mpi3mr_stgt_priv_data *stgt_priv_data;
	struct mpi3mr_sdev_priv_data *sdev_priv_data;
	struct scsi_device *sdev;
	unsigned long flags, r;

	if ((mrioc->unrecoverable) || (mrioc->reset_in_progress))
		return;

	init_waitqueue_head(&mrioc->sysfs_pending_tm_wq);
	atomic_set(&mrioc->sysfs_tm_pending, 0);
	mrioc->sysfs_tm_issued = 0;
	mrioc->sysfs_tm_terminated_io_count = 0;

	scsi_block_requests(mrioc->shost);

	switch (tm_type) {
	case MPI3_SCSITASKMGMT_TASKTYPE_ABORT_TASK:
		blk_mq_tagset_busy_iter(&mrioc->shost->tag_set,
				mpi3mr_app_issue_abort_task, (void *)mrioc);
		break;

	case MPI3_SCSITASKMGMT_TASKTYPE_TARGET_RESET:
		spin_lock_irqsave(&mrioc->tgtdev_lock, flags);
		list_for_each_entry(tgtdev, &mrioc->tgtdev_list, list)
			if (mrioc->sysfs_tm_issued < MPI3MR_NUM_SYSFS_TM)
				mpi3mr_app_issue_tm(mrioc, tm_type,
				    tgtdev->dev_handle, 0, NULL);
		spin_unlock_irqrestore(&mrioc->tgtdev_lock, flags);
		break;

	case MPI3_SCSITASKMGMT_TASKTYPE_LOGICAL_UNIT_RESET:
	case MPI3_SCSITASKMGMT_TASKTYPE_ABORT_TASK_SET:
		shost_for_each_device(sdev, mrioc->shost) {
			 /* wait for free hpr message frames */
			sdev_priv_data = sdev->hostdata;
			if (!sdev_priv_data ||
			    !sdev_priv_data->tgt_priv_data)
				continue;
			stgt_priv_data = sdev_priv_data->tgt_priv_data;
			if (mrioc->sysfs_tm_issued < MPI3MR_NUM_SYSFS_TM)
				mpi3mr_app_issue_tm(mrioc, tm_type,
				    stgt_priv_data->dev_handle,
				    sdev_priv_data->lun_id, NULL);
		}
		break;
	}
	scsi_unblock_requests(mrioc->shost);

	if (atomic_read(&mrioc->sysfs_tm_pending)) {
		r = wait_event_timeout(mrioc->sysfs_pending_tm_wq,
		    !atomic_read(&mrioc->sysfs_tm_pending),
		    MPI3MR_SYSFS_TM_TIMEOUT*HZ);
		if (!r) {
			ioc_err(mrioc,
			    "%s: %d TM requests timed out\n", __func__,
			    atomic_read(&mrioc->sysfs_tm_pending));
			mpi3mr_soft_reset_handler(mrioc,
			    MPI3MR_RESET_FROM_SYSFS_TIMEOUT, 1);
		}
	}

	ioc_info(mrioc, "%s: task management requests issued(%d)\n", __func__,
	    mrioc->sysfs_tm_issued);
	ioc_info(mrioc, "%s: number of IOs terminated(%d)\n", __func__,
	    mrioc->sysfs_tm_terminated_io_count);
}

/**
 * mpi3mr_app_task_management_store- Issue a TM/controller reset
 * @cdev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 * @count: size of the buffer
 *
 * A sysfs 'read/write' shost attribute, to issue a set of TMs
 * or a controller reset to validate the controller firmware for
 * user applications.
 *
 * Return: strlen() return
 */
static ssize_t
mpi3mr_app_task_management_store(struct device *cdev,
    struct device_attribute *attr, const char *buf, size_t count)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct mpi3mr_ioc *mrioc = shost_priv(shost);
	int opcode = 0;

	if (sscanf(buf, "%d", &opcode) != 1)
		return -EINVAL;
	if (mrioc->unrecoverable)
		return -EINVAL;

	switch (opcode) {

	case MPI3MR_SYSFS_TM_SOFT_RESET:
		scsi_block_requests(mrioc->shost);
		ioc_info(mrioc, "%s: soft reset issued, status=%s\n", __func__,
		    ((!mpi3mr_soft_reset_handler(mrioc, MPI3MR_RESET_FROM_SYSFS,
			0)) ? "SUCCESS" : "FAILED"));
		scsi_unblock_requests(mrioc->shost);
		break;

	case MPI3MR_SYSFS_TM_DIAG_FAULT_RESET:
		scsi_block_requests(mrioc->shost);
		ioc_info(mrioc, "%s: diag fault reset issued, status=%s\n",
		    __func__,
		    ((!mpi3mr_soft_reset_handler(mrioc, MPI3MR_RESET_FROM_SYSFS,
			1)) ? "SUCCESS" : "FAILED"));
		scsi_unblock_requests(mrioc->shost);
		break;

	case MPI3MR_SYSFS_TM_ABORT_TASK:
		ioc_info(mrioc, "%s: abort task issued\n", __func__);
		mpi3mr_app_tm_sysfs(mrioc,
		    MPI3_SCSITASKMGMT_TASKTYPE_ABORT_TASK);
		break;

	case MPI3MR_SYSFS_TM_TARGET_RESET:
		ioc_info(mrioc, "%s: target reset issued\n", __func__);
		mpi3mr_app_tm_sysfs(mrioc,
		    MPI3_SCSITASKMGMT_TASKTYPE_TARGET_RESET);
		break;

	case MPI3MR_SYSFS_TM_LUN_RESET:
		ioc_info(mrioc, "%s: lun reset issued\n", __func__);
		mpi3mr_app_tm_sysfs(mrioc,
		    MPI3_SCSITASKMGMT_TASKTYPE_LOGICAL_UNIT_RESET);
		break;

	case MPI3MR_SYSFS_TM_ABORT_TASK_SET:
		ioc_info(mrioc, "%s: abort task set issued\n", __func__);
		mpi3mr_app_tm_sysfs(mrioc,
		    MPI3_SCSITASKMGMT_TASKTYPE_ABORT_TASK_SET);
		break;

	default:
		ioc_warn(mrioc, "%s: unsupported opcode(%d)\n",
		    __func__, opcode);
		break;
	};

	return strlen(buf);
}
static DEVICE_ATTR(task_management, 0200, NULL,
	mpi3mr_app_task_management_store);

#if (KERNEL_VERSION(5, 16, 0) > LINUX_VERSION_CODE)
struct device_attribute *mpi3mr_host_attrs[] = {
	&dev_attr_version_fw,
	&dev_attr_fw_queue_depth,
	&dev_attr_op_req_q_count,
	&dev_attr_reply_queue_count,
	&dev_attr_logging_level,
	&dev_attr_task_management,
	NULL,
};
#else
static struct attribute *mpi3mr_host_attrs[] = {
	&dev_attr_version_fw.attr,
	&dev_attr_fw_queue_depth.attr,
	&dev_attr_op_req_q_count.attr,
	&dev_attr_reply_queue_count.attr,
	&dev_attr_logging_level.attr,
	&dev_attr_task_management.attr,
	NULL,
};

static const struct attribute_group mpi3mr_host_attr_group = {
	.attrs = mpi3mr_host_attrs
};

const struct attribute_group *mpi3mr_host_groups[] = {
	&mpi3mr_host_attr_group,
	NULL,
};
#endif

/*
 * SCSI Device attributes under sysfs
 */

/**
 * mpi3mr_scsih_ncq_prio_supp - Check ncq priority is supported
 * @sdev: scsi device struct
 *
 * This function returns whether the given sdev is capable for
 * setting NCQ priority or not.
 *
 * Return: 0 when device doesn't support NCQ, 1 otherwise
 */
u8 mpi3mr_scsih_ncq_prio_supp(struct scsi_device *sdev)
{
	unsigned char *buf;
	u8 ncq_prio_supp = 0;

	if (!scsi_device_supports_vpd(sdev))
		return ncq_prio_supp;

	buf = kmalloc(SCSI_VPD_PG_LEN, GFP_KERNEL);
	if (!buf)
		return ncq_prio_supp;

	if (!scsi_get_vpd_page(sdev, 0x89, buf, SCSI_VPD_PG_LEN))
		ncq_prio_supp = (buf[213] >> 4) & 1;

	kfree(buf);
	return ncq_prio_supp;
}

/**
 * mpi3mr_app_device_ncq_prio_enable_show - NCQ priority value
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 *
 * A sysfs 'read/write' sdev attribute, to display NCQ priority
 * value. only works with SATA.
 *
 * Return: snprintf() return
 */
static ssize_t
mpi3mr_app_device_ncq_prio_enable_show(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct scsi_device *sdev = to_scsi_device(dev);
	struct mpi3mr_sdev_priv_data *sdev_priv_data = sdev->hostdata;

	return snprintf(buf, PAGE_SIZE, "%d\n",
			sdev_priv_data->ncq_prio_enable);
}

/**
 * mpi3mr_app_device_ncq_prio_enable_store - NCQ priority change
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 * @count: size of the buffer
 *
 * A sysfs 'read/write' sdev attribute, to store NCQ priority
 * value. only works with SATA.
 *
 * Return: strlen() return
 */
static ssize_t
mpi3mr_app_device_ncq_prio_enable_store(struct device *dev,
				struct device_attribute *attr,
				const char *buf, size_t count)
{
	struct scsi_device *sdev = to_scsi_device(dev);
	struct mpi3mr_sdev_priv_data *sdev_priv_data = sdev->hostdata;
	int ncq_prio_enable = 0;

	if (sscanf(buf, "%d", &ncq_prio_enable) != 1)
		return -EINVAL;

	if (!mpi3mr_scsih_ncq_prio_supp(sdev))
		return -EINVAL;

	sdev_priv_data->ncq_prio_enable = ncq_prio_enable;
	return strlen(buf);
}
static DEVICE_ATTR(sata_ncq_prio_enable, 0644,
			mpi3mr_app_device_ncq_prio_enable_show,
			mpi3mr_app_device_ncq_prio_enable_store);

/**
 * sas_address_show - SysFS callback for dev SASaddress display
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 *
 * Return: snprintf() return after copying SAS address of the
 * specific SAS/SATA end device.
 */
static ssize_t
sas_address_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct scsi_device *sdev = to_scsi_device(dev);
	struct mpi3mr_sdev_priv_data *sdev_priv_data;
	struct mpi3mr_stgt_priv_data *tgt_priv_data;
	struct mpi3mr_tgt_dev *tgtdev;

	sdev_priv_data = sdev->hostdata;
	if (!sdev_priv_data)
		return 0;

	tgt_priv_data = sdev_priv_data->tgt_priv_data;
	if (!tgt_priv_data)
		return 0;
	tgtdev = tgt_priv_data->tgt_dev;
	if (!tgtdev || tgtdev->dev_type != MPI3_DEVICE_DEVFORM_SAS_SATA)
		return 0;
	return snprintf(buf, PAGE_SIZE, "0x%016llx\n",
	    (unsigned long long)tgtdev->dev_spec.sas_sata_inf.sas_address);
}

static DEVICE_ATTR_RO(sas_address);

/**
 * device_handle_show - SysFS callback for device handle display
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 *
 * Return: snprintf() return after copying firmware internal
 * device handle of the specific device.
 */
static ssize_t
device_handle_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct scsi_device *sdev = to_scsi_device(dev);
	struct mpi3mr_sdev_priv_data *sdev_priv_data;
	struct mpi3mr_stgt_priv_data *tgt_priv_data;
	struct mpi3mr_tgt_dev *tgtdev;

	sdev_priv_data = sdev->hostdata;
	if (!sdev_priv_data)
		return 0;

	tgt_priv_data = sdev_priv_data->tgt_priv_data;
	if (!tgt_priv_data)
		return 0;
	tgtdev = tgt_priv_data->tgt_dev;
	if (!tgtdev)
		return 0;
	return snprintf(buf, PAGE_SIZE, "0x%04x\n", tgtdev->dev_handle);
}

static DEVICE_ATTR_RO(device_handle);

/**
 * persistent_id_show - SysFS callback for persisten ID display
 * @dev: class device
 * @attr: Device attributes
 * @buf: Buffer to copy
 *
 * Return: snprintf() return after copying persistent ID of the
 * of the specific device.
 */
static ssize_t
persistent_id_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct scsi_device *sdev = to_scsi_device(dev);
	struct mpi3mr_sdev_priv_data *sdev_priv_data;
	struct mpi3mr_stgt_priv_data *tgt_priv_data;
	struct mpi3mr_tgt_dev *tgtdev;

	sdev_priv_data = sdev->hostdata;
	if (!sdev_priv_data)
		return 0;

	tgt_priv_data = sdev_priv_data->tgt_priv_data;
	if (!tgt_priv_data)
		return 0;
	tgtdev = tgt_priv_data->tgt_dev;
	if (!tgtdev)
		return 0;
	return snprintf(buf, PAGE_SIZE, "%d\n", tgtdev->perst_id);
}
static DEVICE_ATTR_RO(persistent_id);

#if (KERNEL_VERSION(5, 16, 0) > LINUX_VERSION_CODE)
struct device_attribute *mpi3mr_dev_attrs[] = {
	&dev_attr_sata_ncq_prio_enable,
	&dev_attr_sas_address,
	&dev_attr_device_handle,
	&dev_attr_persistent_id,
	NULL,
};
#else
static struct attribute *mpi3mr_dev_attrs[] = {
	&dev_attr_sata_ncq_prio_enable.attr,
	&dev_attr_sas_address.attr,
	&dev_attr_device_handle.attr,
	&dev_attr_persistent_id.attr,
	NULL,
};

static const struct attribute_group mpi3mr_dev_attr_group = {
	.attrs = mpi3mr_dev_attrs
};

const struct attribute_group *mpi3mr_dev_groups[] = {
	&mpi3mr_dev_attr_group,
	NULL,
};
#endif

